"""
Resource and ResourceProperty class definitions.
Resource is the base building block of the eiko language model.
"""
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Union

from .base_types import EikoBaseType, EikoPromise, EikoResource, EikoType

if TYPE_CHECKING:
    from ...handlers import Handler
    from .._parser import ResourceDefinitionAST, TypeExprAST
    from .._token import Token
    from .base_model import EikoBaseModel
    from .function import ConstructorDefinition


@dataclass
class ResourceProperty:
    """
    Internal representation of a resource property for constructors.
    It does not verify wether the default value given is the correct type.
    """

    name: str
    type: EikoType
    default_value: Optional[EikoBaseType] = None
    type_expr: "TypeExprAST | None" = None


@dataclass
class EikoPromiseDefinition:
    """
    Internal representation of a promise constructor.
    """

    name: str
    type: EikoType

    def execute(self, callee_token: "Token", parent: EikoResource) -> EikoPromise:
        return EikoPromise(
            self.name,
            self.type,
            callee_token,
            parent,
        )


EikoResourceDefinitionType = EikoType("ResourceDefinition")
PropertiesDict = dict[str, Union[ResourceProperty, EikoPromiseDefinition]]


class EikoResourceDefinition(EikoBaseType):
    """Internal representation of a resource definition."""

    def __init__(
        self,
        expr: "ResourceDefinitionAST",
        default_constructor: "ConstructorDefinition",
        properties: PropertiesDict,
        promises: list[EikoPromiseDefinition],
    ) -> None:
        super().__init__(EikoResourceDefinitionType)
        self.expr = expr
        self.name = expr.name
        self.instance_type: EikoType = expr.type
        self.default_constructor = default_constructor
        self.default_constructor.parent = self
        self.properties: PropertiesDict = properties
        self.index_def = [list(properties.keys())[0]]
        self.promises: list[EikoPromiseDefinition] = promises
        self.handler: Optional[type["Handler"]] = None
        self.linked_basemodel: Optional[type["EikoBaseModel"]] = None

    def printable(self, indent: str = "") -> str:
        return_str = f"{indent}Resource Definition '{self.name}'"
        if self.type.super is not None:
            return_str += f"('{self.type.super.name}')"
        return_str += ": " + "{\n"

        for value in self.properties.values():
            return_str += indent + "    "
            if isinstance(value, EikoPromiseDefinition):
                return_str += "promise "
            return_str += f"{value.name}: {value.type}\n"

        return_str += indent + "}\n"

        return return_str

    def truthiness(self) -> bool:
        raise NotImplementedError



"""
Lexer Tokens describe source content.
They are used to construct the AST by the parser.
"""
from dataclasses import dataclass
from enum import Enum, auto

from .misc import Index


class TokenType(Enum):
    """
    Enum of all potential TokenTypes.
    """

    RESOURCE = auto()
    IMPLEMENT = auto()
    FROM = auto()
    IMPORT = auto()
    TYPEDEF = auto()
    IF = auto()
    ELIF = auto()
    ELSE = auto()
    AND = auto()
    OR = auto()
    DEF = auto()
    PROMISE = auto()
    ENUM = auto()
    FOR = auto()
    IN = auto()

    IDENTIFIER = auto()

    LEFT_PAREN = auto()
    RIGHT_PAREN = auto()
    LEFT_SQ_BRACKET = auto()
    RIGHT_SQ_BRACKET = auto()
    LEFT_BRACE = auto()
    RIGHT_BRACE = auto()
    COLON = auto()
    COMMA = auto()
    DOT = auto()
    DOUBLE_DOT = auto()
    TRIPLE_DOT = auto()
    AT_SIGN = auto()

    DOUBLE_COLON = auto()
    ASSIGNMENT_OP = auto()
    ARITHMETIC_OP = auto()
    COMPARISON_OP = auto()

    # Base types
    INTEGER = auto()
    FLOAT = auto()
    STRING = auto()
    F_STRING = auto()
    TRUE = auto()
    FALSE = auto()

    INDENT = auto()
    EOF = auto()
    UNKNOWN = auto()


token_to_char = {
    TokenType.LEFT_PAREN: "(",
    TokenType.RIGHT_PAREN: ")",
    TokenType.LEFT_SQ_BRACKET: "[",
    TokenType.RIGHT_SQ_BRACKET: "]",
    TokenType.LEFT_BRACE: "{",
    TokenType.RIGHT_BRACE: "}",
    TokenType.COMMA: ",",
    TokenType.DOT: ".",
    TokenType.AT_SIGN: "@",
}


@dataclass
class Token:
    """
    A token represents a gramatical construct
    that can be used to build the AST.
    """

    type: TokenType
    content: str
    index: Index

    def __repr__(self) -> str:
        return f"Token(type=<{self.type.name}>, index=<{self.index.__repr__()}>)"



"""
A project differs from a module in that
a project is meant to be deployed.
It containes modules and a model.
"""
import sys
import tomllib
from pathlib import Path

from packaging import version
from pydantic import BaseModel, ValidationError

from .. import VERSION
from ..core import logger
from .package_manager import install_pkgs, install_py_deps


class ProjectSettings(BaseModel):
    """
    All possible settings read from an eiko.project file
    """

    exists: bool = False

    entry_point: str | None = None
    eikobot_version: str | None = None
    eikobot_requires: list[str] = []
    python_requires: list[str] = []
    dry_run: bool = False
    ssh_timeout: int = 3


def read_project() -> ProjectSettings:
    """
    Reads an eiko.toml file in the cwd and returns it as a settings object.
    """
    path = Path("eiko.toml")
    if not path.exists():
        return ProjectSettings()

    logger.debug("Reading eiko toml project file.")
    try:
        toml = tomllib.loads(Path("eiko.toml").read_text(encoding="utf-8"))
        project_toml = toml.get("eiko", {}).get("project")
        if project_toml is None:
            logger.debug("No project settings.")
            return ProjectSettings()

        return ProjectSettings(**project_toml, exists=True)
    except (tomllib.TOMLDecodeError, ValidationError) as e:
        logger.error(f"Failed to parse eiko.toml: {e}")
        sys.exit(1)


def eiko_version_match(  # pylint: disable=too-many-branches
    eikobot_version: str,
) -> bool:
    """
    Checks to see if the library is compatible with
    the currently installed version of eikobot.
    """
    if eikobot_version is None:
        return True

    eiko_version = version.parse(VERSION)
    for version_req in eikobot_version.split(","):
        if version_req.startswith(">="):
            if eiko_version < _parse_version(version_req[2:]):
                return False

        elif version_req.startswith("<="):
            if eiko_version > _parse_version(version_req[2:]):
                return False

        elif version_req.startswith("=="):
            if eiko_version != _parse_version(version_req[2:]):
                return False

        elif version_req.startswith("!="):
            if eiko_version == _parse_version(version_req[2:]):
                return False

        elif version_req.startswith(">"):
            if eiko_version <= _parse_version(version_req[1:]):
                return False

        elif version_req.startswith("<"):
            if eiko_version >= _parse_version(version_req[1:]):
                return False

        else:
            raise ValueError(
                f"Failed to parse eiko.toml eikobot_version option '{eikobot_version}'."
            )

    return True


def _parse_version(_version: str) -> version.Version:
    try:
        return version.parse(_version)
    except version.InvalidVersion as e:
        raise ValueError(f"failed to parse version `{_version}`.") from e


PROJECT_SETTINGS = read_project()
if PROJECT_SETTINGS.eikobot_version:
    try:
        if not eiko_version_match(PROJECT_SETTINGS.eikobot_version):
            logger.error(
                "Eikobot is not a matching version! "
                "".join(PROJECT_SETTINGS.eikobot_version)
            )
            sys.exit(1)
    except ValueError as err:
        logger.error(str(err))
        sys.exit(1)


async def init_project() -> None:
    """
    Initializes a project based on its eiko.toml settings.
    """
    logger.info("Initializing eikobot project.")
    if PROJECT_SETTINGS.exists:
        logger.info("Installing Project Python requirements.")
        install_py_deps(PROJECT_SETTINGS.python_requires)

        logger.info("Installing Project Eikobot requirements.")
        await install_pkgs(PROJECT_SETTINGS.eikobot_requires)



"""
The nuts and bolts of Eikobot.
"""


def human_readable(number: int) -> str:
    """
    Turns a large integer representing bytes
    in to a human readable string.
    """
    number = number // 8
    for unit in ["", "K", "M", "G", "T"]:
        if number < 10240:
            return f"{number}{unit}B"
        number = number // 1024
    return f"{number}PB"


def machine_readable(number: str) -> int:
    """
    Turns a human readable number of bytes in to an integer.
    """
    if number.isdigit():
        return int(number)

    if number.endswith("B"):
        return int(number[:-1])

    if number.endswith("KB"):
        return int(number[:-2]) * 1024

    if number.endswith("MB"):
        return int(number[:-2]) * 1024**2

    if number.endswith("GB"):
        return int(number[:-2]) * 1024**3

    if number.endswith("TB"):
        return int(number[:-2]) * 1024**4

    if number.endswith("PB"):
        return int(number[:-2]) * 1024**5

    raise ValueError(f"Cannot convert value of '{number}' to an integer.")



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-statements
from pathlib import Path

import pytest

from eikobot.core.deployer import Deployer
from eikobot.core.exporter import Exporter


@pytest.mark.asyncio
async def test_deploy(eiko_deploy_file: Path) -> None:
    exporter = Exporter()
    exporter.export_from_file(eiko_deploy_file)
    deployer = Deployer()
    await deployer.deploy(exporter)

    base_handler_1 = exporter.base_tasks[0].handler
    base_handler_2 = exporter.base_tasks[1].handler
    base_handler_3 = exporter.base_tasks[2].handler

    # We can't actually easily get these types,
    # so we just ignore the typing here.
    assert base_handler_1.created  # type: ignore
    assert base_handler_2.created  # type: ignore
    assert base_handler_3.created  # type: ignore
    assert base_handler_1.create_called == 1  # type: ignore
    assert base_handler_2.create_called == 1  # type: ignore
    assert base_handler_3.create_called == 1  # type: ignore
    assert base_handler_1.update_called == 0  # type: ignore
    assert base_handler_2.update_called == 0  # type: ignore
    assert base_handler_3.update_called == 0  # type: ignore

    mid_handler_1 = exporter.base_tasks[0].dependants[0].handler
    mid_handler_2 = exporter.base_tasks[1].dependants[0].handler
    mid_handler_3 = exporter.base_tasks[2].dependants[0].handler

    assert mid_handler_1.created  # type: ignore
    assert mid_handler_2.created  # type: ignore
    assert mid_handler_3.created  # type: ignore
    assert mid_handler_1.create_called == 1  # type: ignore
    assert mid_handler_2.create_called == 1  # type: ignore
    assert mid_handler_3.create_called == 1  # type: ignore
    assert mid_handler_1.update_called == 0  # type: ignore
    assert mid_handler_2.update_called == 0  # type: ignore
    assert mid_handler_3.update_called == 0  # type: ignore

    top_handler_1 = exporter.base_tasks[0].dependants[0].dependants[0].handler
    top_handler_2 = exporter.base_tasks[0].dependants[0].dependants[1].handler

    assert top_handler_1.created  # type: ignore
    assert top_handler_2.created  # type: ignore
    assert top_handler_1.create_called == 1  # type: ignore
    assert top_handler_2.create_called == 1  # type: ignore
    assert top_handler_1.update_called == 0  # type: ignore
    assert top_handler_2.update_called == 0  # type: ignore

    await deployer.deploy(exporter)

    assert base_handler_1.created  # type: ignore
    assert base_handler_2.created  # type: ignore
    assert base_handler_3.created  # type: ignore
    assert base_handler_1.create_called == 2  # type: ignore
    assert base_handler_2.create_called == 2  # type: ignore
    assert base_handler_3.create_called == 2  # type: ignore
    assert base_handler_1.update_called == 0  # type: ignore
    assert base_handler_2.update_called == 0  # type: ignore
    assert base_handler_3.update_called == 0  # type: ignore

    mid_handler_1 = exporter.base_tasks[0].dependants[0].handler
    mid_handler_2 = exporter.base_tasks[1].dependants[0].handler
    mid_handler_3 = exporter.base_tasks[2].dependants[0].handler

    assert mid_handler_1.created  # type: ignore
    assert mid_handler_2.created  # type: ignore
    assert mid_handler_3.created  # type: ignore
    assert mid_handler_1.create_called == 1  # type: ignore
    assert mid_handler_2.create_called == 1  # type: ignore
    assert mid_handler_3.create_called == 1  # type: ignore
    assert mid_handler_1.update_called == 0  # type: ignore
    assert mid_handler_2.update_called == 0  # type: ignore
    assert mid_handler_3.update_called == 0  # type: ignore

    top_handler_1 = exporter.base_tasks[0].dependants[0].dependants[0].handler
    top_handler_2 = exporter.base_tasks[0].dependants[0].dependants[1].handler

    assert top_handler_1.created  # type: ignore
    assert top_handler_2.created  # type: ignore
    assert top_handler_1.create_called == 1  # type: ignore
    assert top_handler_2.create_called == 1  # type: ignore
    assert top_handler_1.update_called == 1  # type: ignore
    assert top_handler_2.update_called == 1  # type: ignore

    await deployer.deploy(exporter)

    assert base_handler_1.created  # type: ignore
    assert base_handler_2.created  # type: ignore
    assert base_handler_3.created  # type: ignore
    assert base_handler_1.create_called == 3  # type: ignore
    assert base_handler_2.create_called == 3  # type: ignore
    assert base_handler_3.create_called == 3  # type: ignore
    assert base_handler_1.update_called == 0  # type: ignore
    assert base_handler_2.update_called == 0  # type: ignore
    assert base_handler_3.update_called == 0  # type: ignore

    mid_handler_1 = exporter.base_tasks[0].dependants[0].handler
    mid_handler_2 = exporter.base_tasks[1].dependants[0].handler
    mid_handler_3 = exporter.base_tasks[2].dependants[0].handler

    assert mid_handler_1.created  # type: ignore
    assert mid_handler_2.created  # type: ignore
    assert mid_handler_3.created  # type: ignore
    assert mid_handler_1.create_called == 1  # type: ignore
    assert mid_handler_2.create_called == 1  # type: ignore
    assert mid_handler_3.create_called == 1  # type: ignore
    assert mid_handler_1.update_called == 0  # type: ignore
    assert mid_handler_2.update_called == 0  # type: ignore
    assert mid_handler_3.update_called == 0  # type: ignore

    top_handler_1 = exporter.base_tasks[0].dependants[0].dependants[0].handler
    top_handler_2 = exporter.base_tasks[0].dependants[0].dependants[1].handler

    assert top_handler_1.created  # type: ignore
    assert top_handler_2.created  # type: ignore
    assert top_handler_1.create_called == 1  # type: ignore
    assert top_handler_2.create_called == 1  # type: ignore
    assert top_handler_1.update_called == 2  # type: ignore
    assert top_handler_2.update_called == 2  # type: ignore



"""
The deployer takes the output of the exporter
and goes through the tasks of deploying.
"""
import asyncio
from dataclasses import dataclass
from pathlib import Path

from . import logger
from .exporter import Exporter, Task


@dataclass
class DeployProgress:
    """
    Helper class for deployer to display how far along it is.
    """

    total: int
    log: bool = False
    done: int = 0


class Deployer:
    """
    The deployer takes the output of the exporter
    and goes through the tasks of deploying.
    """

    def __init__(self) -> None:
        self.asyncio_tasks: list[asyncio.Task] = []
        self.log_progress = False
        self.progress = DeployProgress(0)
        self.failed = False

    async def deploy(self, exporter: Exporter, log_progress: bool = False) -> None:
        """
        Given a set of Tasks, walks through them and makes sure they're all done.
        """
        self.failed = False
        self.log_progress = log_progress
        self.progress = DeployProgress(exporter.total_tasks, log_progress)
        for task in exporter.base_tasks:
            task.init(self._done_cb, self._failure_cb)
            self._create_task(task)

        while self.asyncio_tasks:
            await asyncio.gather(*self.asyncio_tasks)

        logger.info("Cleaning up.")
        for task in exporter.task_index.values():
            if task.handler is not None:
                await task.handler.cleanup(task.ctx)

    async def dry_run(self, exporter: Exporter) -> None:
        """Executes a dry run of all tasks."""
        for task in exporter.base_tasks:
            task.init()

        for task in exporter.task_index.values():
            if task.handler is not None:
                task.ctx.resource = task.ctx.raw_resource.to_py()
                await task.handler.__dry_run__(task.ctx)

    async def deploy_from_file(self, eiko_file: Path) -> None:
        """Helper funcion meant mostly for testing."""
        exporter = Exporter()
        exporter.export_from_file(eiko_file)
        await self.deploy(exporter)

    def _create_task(self, task: Task) -> None:
        asyncio_task = asyncio.create_task(self._execute_task(task))
        self.asyncio_tasks.append(asyncio_task)

    async def _execute_task(self, task: Task) -> None:
        await task.execute()
        for sub_task in task.dependants:
            if not sub_task.depends_on_copy:
                self._create_task(sub_task)

        asyncio_task = asyncio.current_task()
        if asyncio_task is not None:
            self.asyncio_tasks.remove(asyncio_task)

    def _done_cb(self) -> None:
        self.progress.done += 1
        if self.progress.log:
            logger.info(f"{self.progress.done} of {self.progress.total} tasks done.")

    def _failure_cb(self) -> None:
        self.failed = True



"""
The exporter takes the output from the compiler
and turns it in tasks the deployer understands.
"""
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional, Union

from . import logger
from .compiler import Compiler, CompilerContext
from .errors import EikoDeployError, EikoExportError, EikoInternalError
from .handlers import Handler, HandlerContext
from .helpers import EikoDict, EikoList, EikoResource


@dataclass
class Task:
    """A task is a piece of work the backend needs to do."""

    task_id: str
    ctx: HandlerContext
    handler: Optional[Handler]

    def __post_init__(self) -> None:
        self.dependants: list["Task"] = []
        self.depends_on: list["Task"] = []
        self.depends_on_copy: list["Task"] = []
        self._done_cb: Optional[Callable[[], None]] = None
        self._failure_cb: Optional[Callable[[], None]] = None

    def init(
        self,
        done_cb: Optional[Callable[[], None]] = None,
        failure_cb: Optional[Callable[[], None]] = None,
    ) -> None:
        """Resets a task and it's sub tasks so they can run again."""
        self._done_cb = done_cb
        self._failure_cb = failure_cb
        self.depends_on_copy = self.depends_on.copy()
        for dependant in self.dependants:
            dependant.init(done_cb, failure_cb)

    async def execute(self) -> None:
        """Executes the task, than let's it's dependants know it's done."""
        logger.info(f"Starting task '{self.task_id}'")
        self.ctx.resource = self.ctx.raw_resource.to_py()
        await self._run_handler()

        if self.ctx.failed or not self.ctx.deployed:
            if self._failure_cb is not None:
                self._failure_cb()
            return

        for name, promise in self.ctx.raw_resource.promises.items():
            if promise.value is None:
                logger.error(
                    f"Resource '{self.task_id}' was deployed, "
                    f"but promise '{name}' was not fullfilled."
                )
                if self._failure_cb is not None:
                    self._failure_cb()
                return

        logger.debug(f"Done executing task '{self.task_id}'")
        for sub_task in self.dependants:
            sub_task.remove_dep(self)

        if self._done_cb is not None:
            self._done_cb()

    async def _run_handler(self) -> None:
        if self.handler is not None:
            try:
                await self.handler.__pre__(self.ctx)
                if self.ctx.failed:
                    raise EikoDeployError("Pre deploy failed.")

                await self.handler.execute(self.ctx)
                if self.ctx.failed or not self.ctx.deployed:
                    raise EikoDeployError("Handler execution failed.")

                await self.handler.resolve_promises(self.ctx)
                if self.ctx.failed:
                    raise EikoDeployError("Resolving promises failed.")

            except Exception as e:  # pylint: disable=broad-exception-caught
                self.ctx.failed = True
                logger.error(f"Failed to deploy '{self.task_id}': {e}")

            try:
                await self.handler.__post__(self.ctx)
            except Exception as e:
                raise EikoDeployError(
                    f"Failed to cleanup for task '{self.task_id}': {e}",
                ) from e

        else:
            raise EikoInternalError(
                "Deployer failed to execute a task because a handler was missing. "
            )

    def remove_dep(self, task: "Task") -> None:
        self.depends_on_copy.remove(task)

    def process_sub_task(self, sub_task: "Task") -> None:
        """
        Adds a task as a dependancy for this task.
        Sets up in such a way that dependancies are correct.
        """
        if sub_task.handler is not None:
            self.depends_on.append(sub_task)
            if self.handler is not None:
                sub_task.add_dependant(self)
        else:
            for sub_sub_task in sub_task.depends_on:
                self.add_depends_on(sub_sub_task)
                if self.handler is not None:
                    sub_sub_task.add_dependant(self)

    def add_dependant(self, task: "Task") -> None:
        if task not in self.dependants:
            self.dependants.append(task)

    def add_depends_on(self, task: "Task") -> None:
        if task not in self.depends_on:
            self.depends_on.append(task)

    def process_promise(self, super_task: "Task") -> None:
        """
        Calculates dependencies for a task that depends on a promise.

        The passed super_task should be the one releasing the promise.
        """
        if super_task.handler is None:
            raise EikoExportError(
                f"Resource '{super_task.ctx.raw_resource.type.name}' "
                "has promises but no handler.",
            )

        self.process_sub_task(super_task)


class Exporter:
    """
    The exporter takes the output from the compiler
    and turns it in tasks the deployer understands.
    """

    def __init__(self) -> None:
        self.task_index: dict[str, Task] = {}
        self.base_tasks: list[Task] = []
        self.total_tasks: int = 0

    def export_from_file(self, file: Path) -> None:
        """Compiles a file and exports the tasks."""
        logger.debug("Constructing task dependency trees.")
        compiler = Compiler()
        compiler.compile(file)
        self.export_from_context(compiler.context)

    def export_from_context(self, context: CompilerContext) -> None:
        """
        Walks through a compiler context and exports it as a set fo tasks.
        """
        values = list(context.storage.values())
        values.extend(context.orphans)

        for value in values:
            if isinstance(value, EikoResource):
                self._parse_task(value)
            elif isinstance(value, (EikoList, EikoDict)):
                self._parse_multi(value)

    def _parse_multi(self, value: Union[EikoList, EikoDict]) -> list[Task]:
        tasks: list[Task] = []
        if isinstance(value, EikoList):
            item_list = value.elements

        elif isinstance(value, EikoDict):
            item_list = list(value.elements.values())

        for item in item_list:
            if isinstance(item, EikoResource):
                tasks.append(self._parse_task(item))
            elif isinstance(item, (EikoList, EikoDict)):
                tasks.extend(self._parse_multi(item))

        return tasks

    def _parse_task(self, resource: EikoResource) -> Task:
        task_id = resource.index()
        pre_task = self.task_index.get(task_id)
        if pre_task is not None:
            return pre_task

        handler = None
        if resource.class_ref.handler is not None:
            handler = resource.class_ref.handler()
            self.total_tasks += 1

        _id = resource.index()
        task = Task(
            _id,
            HandlerContext(resource, _id),
            handler,
        )

        for value in resource.properties.values():
            if isinstance(value, EikoResource):
                task.process_sub_task(self._parse_task(value))
            elif isinstance(value, (EikoList, EikoDict)):
                for item in self._parse_multi(value):
                    task.process_sub_task(item)

        for name, promise in resource.get_external_promises():
            super_task = self.task_index.get(promise.parent.index())
            if super_task is None:
                raise EikoExportError(
                    f"Task '{task_id}' depends on promise "
                    f"'{promise.parent.index()}.{name}', but this promise has no associated task."
                )
            task.process_promise(super_task)

        self.task_index[task_id] = task
        if task.handler is not None and not task.depends_on:
            self.base_tasks.append(task)

        return task



"""
While real functions don't exist in the eiko language,
constructors and plugins do, and they need some kind of representation.
"""
import inspect
from dataclasses import dataclass
from pathlib import Path
from types import UnionType
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Type,
    Union,
    get_args,
    get_origin,
)

from ...errors import EikoCompilationError, EikoInternalError, EikoPluginError
from ...plugin import EikoPluginException, EikoPluginTyping
from .._token import Token
from .base_types import (
    INDEXABLE_TYPES,
    EikoBaseType,
    EikoProtectedStr,
    EikoResource,
    EikoType,
    EikoUnset,
    PassedArg,
    PyTypes,
    to_eiko,
    to_eiko_type,
)

if TYPE_CHECKING:
    from .._parser import ExprAST
    from ._resource import EikoResourceDefinition
    from .context import CompilerContext, StorableTypes

EikoFunctionType = EikoType("function")


@dataclass
class ConstructorArg:
    """Representation of a required constructor argument."""

    name: str
    type: EikoType
    default_value: Optional[EikoBaseType] = None


class ConstructorDefinition(EikoBaseType):
    """Internal representation of an Eikobot constructor."""

    def __init__(self, name: str, execution_context: "CompilerContext") -> None:
        super().__init__(EikoFunctionType)
        self.parent: "EikoResourceDefinition"
        self.name = name
        self.args: dict[str, ConstructorArg] = {}
        self.body: list["ExprAST"] = []
        self.execution_context = execution_context
        self.index_def: list[str] = []

    def printable(self, _: str = "") -> str:
        raise NotImplementedError

    def add_arg(self, arg: ConstructorArg) -> None:
        self.args[arg.name] = arg

    def add_body_expr(self, expr: "ExprAST") -> None:
        self.body.append(expr)

    def execute(  # pylint: disable=too-many-locals,too-many-branches
        self,
        callee_token: Token,
        positional_args: list[PassedArg],
        keyword_args: dict[str, PassedArg],
    ) -> EikoResource:
        """
        Executes a function call based on a premade constructor spec.
        """
        if self.parent.promises and self.parent.handler is None:
            raise EikoCompilationError(
                f"Resource Definition '{self.parent.name}' has promises, "
                "but no handler.",
                token=self.parent.expr.token,
            )

        if len(positional_args) + len(keyword_args) > len(self.args) - 1:
            raise EikoCompilationError(
                "Too many arguments given to function call. "
                f"Expected {len(self.args) - 1}, "
                f"got {len(positional_args) + len(keyword_args)}.",
                token=callee_token,
            )

        handled_args: dict[str, EikoBaseType] = {}
        self_arg = list(self.args.values())[0]
        resource = EikoResource(self.parent)
        handled_args[self_arg.name] = resource
        self._handle_args(handled_args, positional_args, keyword_args)

        for arg_name, arg in self.args.items():
            if arg_name not in handled_args:
                if arg.default_value is None:
                    raise EikoCompilationError(
                        f"Argument '{arg_name}' for callable '{callee_token.content}' requires a value.",
                        token=callee_token,
                    )

                handled_args[arg_name] = arg.default_value

        context = self.execution_context.get_subcontext(
            f"{self.parent.name}.{self.name}"
        )
        for prop in self.parent.properties.values():
            if prop in self.parent.promises:
                # prop is garantueed to be an EikoPromiseDef here
                resource.add_promise(prop.execute(callee_token, resource))  # type: ignore
            else:
                resource.populate_property(prop.name, prop.type)

        for arg_name, value in handled_args.items():
            context.set(arg_name, value)

        for expr in self.body:
            expr.compile(context)

        for prop in self.parent.properties.values():
            res_prop = resource.properties.get(prop.name)
            if isinstance(res_prop, EikoUnset):
                raise EikoCompilationError(
                    f"Property '{self.parent.name}.{prop.name}' was not set in the constructor.",
                    token=callee_token,
                )

        res_index = self.parent.name
        for property_name in self.index_def:
            if property_name == self.parent.name:
                continue
            prop_name_split = property_name.split(".")
            index_prop: Union[EikoBaseType, EikoUnset, None] = resource
            for prop_name in prop_name_split:
                if isinstance(index_prop, EikoResource):
                    index_prop = index_prop.properties.get(prop_name)
                else:
                    raise EikoCompilationError(
                        f"Failed to create index using property '{property_name}': "
                        f"No property '{prop_name}'.",
                        token=self.parent.expr.token,
                    )

            if not isinstance(index_prop, INDEXABLE_TYPES):
                raise EikoCompilationError(
                    f"Property '{property_name}' of '{self.parent.name}' is not an indexable type.",
                    token=self.parent.expr.token,
                )

            if isinstance(index_prop, EikoProtectedStr):
                raise EikoCompilationError(
                    f"Property '{property_name}' of '{self.parent.name}' is a protected string. "
                    "It can not be used to create an index.",
                    token=self.parent.expr.token,
                )

            res_index += "-" + index_prop.index()

        resource.set_index(res_index)
        if resource.index() in context.global_id_list:
            raise EikoCompilationError(
                f"A resource of type '{self.parent.name}' with index "
                f"'{resource.index()}' was already created.",
                token=callee_token,
            )

        context.global_id_list.append(resource.index())

        return resource

    def _handle_args(  # pylint: disable=too-many-branches
        self,
        handled_args: dict[str, EikoBaseType],
        positional_args: list[PassedArg],
        keyword_args: dict[str, PassedArg],
    ) -> None:
        for passed_arg, arg in zip(positional_args, list(self.args.values())[1:]):
            if not passed_arg.value.type.type_check(arg.type):
                # Try to coerce the type
                if passed_arg.value.type.inverse_type_check(arg.type):
                    if arg.type.typedef is None:
                        raise EikoInternalError(
                            "Failed to coerce type.",
                            token=passed_arg.token,
                        )
                    passed_arg.value = arg.type.typedef.execute(
                        passed_arg.value, passed_arg.token
                    )
                else:
                    raise EikoCompilationError(
                        f"Argument '{arg.name}' expected value of type '{arg.type}', "
                        f"but got value of type '{passed_arg.value.type}'.",
                        token=passed_arg.token,
                    )

            handled_args[arg.name] = passed_arg.value

        for arg_name, passed_arg in keyword_args.items():
            positional_value = handled_args.get(arg_name)
            if positional_value is not None:
                raise EikoCompilationError(
                    f"'{self.name}()' got multiple values for argument '{arg_name}'."
                )

            kw_arg = self.args.get(arg_name)
            if kw_arg is None:
                raise EikoCompilationError(
                    f"'{self.parent.name}.{self.name}()' has no argument '{arg_name}'.",
                    token=passed_arg.token,
                )

            if not passed_arg.value.type.type_check(kw_arg.type):
                if kw_arg.type.inverse_type_check(passed_arg.value.type):
                    if kw_arg.type.typedef is not None:
                        passed_arg.value = kw_arg.type.typedef.execute(
                            passed_arg.value,
                            passed_arg.token,
                        )
                    else:
                        raise EikoInternalError(
                            "Ran in to a typedef that does not have a constructor. "
                            "This is most likely a bug. PLease report this on Github.",
                            token=passed_arg.token,
                        )
                else:
                    raise EikoCompilationError(
                        f"Argument '{kw_arg.name}' expected value of type '{kw_arg.type}', "
                        f"but got value of type '{passed_arg.value.type}'.",
                        token=passed_arg.token,
                    )

            handled_args[arg_name] = passed_arg.value

    def truthiness(self) -> bool:
        raise NotImplementedError


EikoPluginType = EikoType("plugin")


class DefaultValueNotSet:
    """Indicates no default value was set."""

    def __init__(self) -> None:
        pass


@dataclass
class PluginArg:
    """Class representing an argument in a plugin call."""

    name: str
    py_type: Union[Type[EikoBaseType], Type[PyTypes]]
    default_value: Union[PyTypes, DefaultValueNotSet] = DefaultValueNotSet()


class PluginDefinition(EikoBaseType):
    """
    Internal representation of a python plugin
    that can be called from the Eikobot language.
    """

    def __init__(
        self,
        body: EikoPluginTyping,
        return_type: Type[EikoBaseType],
        identifier: str,
        module: str,
    ) -> None:
        super().__init__(EikoPluginType)
        self.body = body
        self.return_type = to_eiko_type(return_type)
        self._body_return_type = return_type
        self.args: list[PluginArg] = []
        self.identifier = identifier
        self.module = module

    def printable(self, _: str = "") -> str:
        return f"Plugin '{self.identifier}'"

    def add_arg(self, arg: PluginArg) -> None:
        self.args.append(arg)

    def execute(
        self, args: list["ExprAST"], context: "CompilerContext", token: Optional[Token]
    ) -> Optional[EikoBaseType]:
        """Execute the stored function and coerces types."""

        stable_args = []
        for i, arg in enumerate(args):
            stable_args.append(self._handle_arg(arg, context, self.args[i]))

        try:
            val = self.body(*stable_args)
        except EikoPluginException as e:
            raise EikoPluginError(str(e), token=token) from e
        except Exception as e:
            raise EikoPluginError(
                f"An unhandled exception occured while executing plugin '{self.identifier}'",
                token=token,
                python_exception=e,
            ) from e

        return to_eiko(val)

    def _handle_arg(
        self, arg: "ExprAST", context: "CompilerContext", required_arg: PluginArg
    ) -> "StorableTypes | PyTypes":
        compiled_arg = arg.compile(context)
        if compiled_arg is None:
            raise EikoCompilationError(
                f"Plugin '{self.module}.{self.identifier}' arg '{required_arg.name}' expects a value "
                f"but expression did not result in a suitable value.",
                token=arg.token,
            )

        converted_arg: "StorableTypes | PyTypes"
        if required_arg.py_type in [None, bool, float, int, str, dict, list, Path]:
            if isinstance(compiled_arg, EikoBaseType):
                try:
                    converted_arg = compiled_arg.to_py()
                except NotImplementedError:
                    converted_arg = compiled_arg
            else:
                converted_arg = compiled_arg

            if not isinstance(converted_arg, required_arg.py_type):
                raise EikoCompilationError(
                    f"Plugin '{self.module}.{self.identifier}' arg '{required_arg.name}' expects an argument "
                    f"of type '{required_arg.py_type.__name__}', but instead got '{compiled_arg.type}'.",
                    token=arg.token,
                )

            return converted_arg

        converted_arg = compiled_arg

        if self._type_check(converted_arg, required_arg.py_type):
            return converted_arg

        raise EikoCompilationError(
            f"Plugin '{self.module}.{self.identifier}' arg '{required_arg.name}' expects an argument "
            f"of type '{required_arg.py_type}', but instead got '{compiled_arg.type}'.",
            token=arg.token,
        )

    def _type_check(self, arg: "StorableTypes | PyTypes", expected: Type) -> bool:
        try:
            return isinstance(arg, expected)
        except TypeError:
            pass

        origin = get_origin(expected)
        if origin in (Union, UnionType):
            for arg_type in get_args(expected):
                if self._type_check(arg, arg_type):
                    return True
            return False

        if origin is (list, List):
            # if not isinstance(arg, list):
            #     return False
            #  args = get_args(expected)
            raise EikoInternalError(
                "Something went horribly wrong during a python runtime type check."
            )

        if origin in (dict, Dict):
            raise EikoInternalError(
                "Something went horribly wrong during a python runtime type check."
            )

        if origin in (type, Type):
            for arg_type in get_args(expected):
                if inspect.isclass(arg) and issubclass(arg, arg_type):
                    return True
            return False

        raise EikoInternalError(
            "Something went horribly wrong during a python runtime type check."
        )

    def truthiness(self) -> bool:
        raise NotImplementedError



"""
The Eiko standard library contains various quality of life
plugins and resources.
"""
import asyncio
import getpass
import os
import re
from dataclasses import dataclass
from ipaddress import IPv4Address, IPv6Address, ip_address
from pathlib import Path
from typing import Optional, Type, Union

import asyncssh
from asyncssh.connection import SSHClientConnection as SSHConnection
from asyncssh.listener import SSHListener
from colorama import Fore
from pydantic import Extra

from eikobot.core.handlers import Handler, HandlerContext
from eikobot.core.helpers import (
    EikoBaseModel,
    EikoBaseType,
    EikoDeployError,
    EikoPromise,
    EikoProtectedStr,
)
from eikobot.core.plugin import eiko_plugin
from eikobot.core.project import PROJECT_SETTINGS


class SSHTimeout(EikoDeployError):
    def __init__(self, *args: object) -> None:
        super().__init__(
            "Failed to connect to host, SSH timed out.",
            *args,
        )


@eiko_plugin()
def debug_msg(msg: str) -> None:
    """Writes a message to the console."""
    print(Fore.BLUE + "DEBUG_MSG" + Fore.RESET, msg)


@eiko_plugin()
def inspect(obj: EikoBaseType) -> None:
    """Pretty print objects to the console."""
    print("INSPECT " + obj.printable())


@eiko_plugin()
def is_ipv4(addr: str) -> bool:
    return _is_ipaddr(addr, IPv4Address)


@eiko_plugin()
def is_ipv6(addr: str) -> bool:
    return _is_ipaddr(addr, IPv6Address)


def _is_ipaddr(addr: str, ip_type: Union[Type[IPv4Address], Type[IPv6Address]]) -> bool:
    try:
        cast_addr = ip_address(addr)
        if isinstance(cast_addr, ip_type):
            return True

    except ValueError:
        pass

    return False


@dataclass
class CmdResult:
    """The result of a command that was run."""

    cmd: str
    returncode: int
    output: str
    ctx: HandlerContext

    def failed(self) -> bool:
        """
        Logs an error and sets the context to failed
        if the command didn't return 0.
        """
        if self.returncode != 0:
            self.ctx.failed = True
            self.ctx.error("Failed to execute " + f'"{self.cmd}"')
            self.ctx.error(f"log:\n{self.output}")
            return True

        return False


class ForwardedPortListener:
    """Represents a forwarded port that needs to be closed."""

    def __init__(
        self, listener: SSHListener, local_port: int, remote_port: int
    ) -> None:
        self.local_port = local_port
        self.remote_port = remote_port
        self.listener = listener
        self.connections = 1


class HostModel(EikoBaseModel):
    """
    Respresents a host to which you can deploy a resource.
    Can execute ssh commands.
    """

    __eiko_resource__ = "Host"

    host: str
    port: int = 22
    username: Optional[str] = None
    password: Optional[str] = None
    ssh_requires_pass: bool = False
    sudo_requires_pass: bool = True

    os_platform: EikoPromise[str]
    os_name: EikoPromise[str]
    os_version: EikoPromise[str]

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow

    def __post_init__(self) -> None:
        self.is_windows_host: bool = False

        self._connection: SSHConnection
        self._con_ref_count: int = 0
        self._connected: asyncio.Event = asyncio.Event()
        self._disconnect_tasks: list[asyncio.Task] = []
        self._forwarded_ports: dict[int, ForwardedPortListener] = {}
        self._forwarded_ports_stop_tasks: list[asyncio.Task] = []

    async def connect(self, ctx: HandlerContext) -> None:
        """Connects or reuses a existing connection."""
        self._con_ref_count += 1
        if self._con_ref_count == 1:
            ctx.debug(f"SSH: Connecting to '{self.host}'.")
            self._connection = await self._create_connection()
            ctx.debug(f"SSH: Connected to '{self.host}'.")
            self._connected.set()

        else:
            ctx.debug(f"SSH: Reusing existing connection to '{self.host}'.")
            await self._connected.wait()

    async def _create_connection(self) -> SSHConnection:
        extra_args: dict[str, str] = {}
        if self.username is not None:
            extra_args["username"] = self.username

        if self.ssh_requires_pass:
            if self.password is not None:
                extra_args["password"] = self.password

        try:
            return await asyncio.wait_for(
                asyncssh.connect(
                    self.host,
                    **extra_args,
                ),
                timeout=PROJECT_SETTINGS.ssh_timeout,
            )
        except asyncio.TimeoutError as e:
            raise SSHTimeout from e
        except asyncssh.HostKeyNotVerifiable as e:
            key_scan = await asyncio.subprocess.create_subprocess_shell(  # pylint: disable=no-member
                f"ssh {self.host} echo",
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )

            _, stderr = await key_scan.communicate()
            if key_scan.returncode != 0:
                raise EikoDeployError(
                    f"Host verification failed \n{stderr.decode()}"
                ) from e
            return await asyncssh.connect(self.host, **extra_args)

    def disconnect(self, ctx: HandlerContext) -> None:
        """Disconnects if nothing else is using the same connection."""
        self._disconnect_tasks.append(
            asyncio.create_task(
                self._disconnect_delay(ctx),
                name=f"disconnect-{self.host}-{self._con_ref_count}",
            )
        )

    async def _disconnect_delay(self, ctx: HandlerContext) -> None:
        await asyncio.sleep(1)
        self._con_ref_count -= 1
        if self._con_ref_count == 0:
            self._connection.close()
            ctx.debug(f"SSH: Disconnected from '{self.host}'.")

    async def wait_until_disconnected(self) -> None:
        await asyncio.gather(*self._forwarded_ports_stop_tasks)
        await asyncio.gather(*self._disconnect_tasks)

    async def scp_to(
        self,
        file_path: Path,
        ctx: HandlerContext,
        destination: Path | str | None = None,
    ) -> None:
        """
        Copies a file from the local host to the remote host
        """
        if not file_path.exists():
            ctx.error(f"Tried to SCP file that doesn't exist: '{file_path}'.")
            raise FileNotFoundError

        extra_args: dict[str, str] = {}
        if self.username is not None:
            extra_args["username"] = self.username

        if self.ssh_requires_pass:
            if self.password is not None:
                extra_args["password"] = self.password

        if destination is None:
            destination = file_path.name

        ctx.debug(f"Copying file '{file_path}' to host.")
        await asyncssh.scp(
            file_path,
            f"{self.host}:{destination}",
            **extra_args,  # type: ignore
        )

    async def scp_from(
        self,
        file_path: Path | str,
        ctx: HandlerContext,
        destination: Path | str,
    ) -> None:
        """
        Copies a file to the local host from the remote host
        """
        extra_args: dict[str, str] = {}
        if self.username is not None:
            extra_args["username"] = self.username

        if self.ssh_requires_pass:
            if self.password is not None:
                extra_args["password"] = self.password

        ctx.debug(f"Copying file '{file_path}' from host to '{destination}'.")
        await asyncssh.scp(
            f"{self.host}:{file_path}",
            destination,
            **extra_args,  # type: ignore
        )

    async def script(
        self, script: str, exec_shell: str, ctx: HandlerContext
    ) -> CmdResult:
        """Runs a script on the remote host."""
        if self.is_windows_host:
            script_file_name = f"script-{ctx.normalized_task_id()}.ps1"
            with open(ctx.task_cache / script_file_name, "w", encoding="utf-8") as f:
                f.write(script)

            extra_args: dict[str, str] = {}
            if self.username is not None:
                extra_args["username"] = self.username

            if self.ssh_requires_pass:
                if self.password is not None:
                    extra_args["password"] = self.password

            await asyncssh.scp(
                ctx.task_cache / script_file_name,
                f"{self.host}:{script_file_name}",
                **extra_args,  # type: ignore
            )
            result = await self._execute_windows(
                f".\\{script_file_name}",
                ctx,
                script,
            )
            await self._connection.run(
                f"del {script_file_name}",
                term_type="xterm-color",
            )
            return result

        if "sudo " in script:
            ssh_exec = self._execute_sudo
        else:
            ssh_exec = self._execute

        _script = f"{exec_shell} << EOF\n{script}\nEOF"
        return await ssh_exec(_script, ctx, script)

    async def execute(self, cmd: str, ctx: HandlerContext) -> CmdResult:
        """Executes one or more commands in an ssh session."""
        if self.is_windows_host:
            ssh_exec = self._execute_windows
        elif "sudo " in cmd:
            ssh_exec = self._execute_sudo
        else:
            ssh_exec = self._execute

        return await ssh_exec(cmd, ctx)

    async def _execute(
        self,
        cmd: str,
        ctx: HandlerContext,
        original_command: Optional[str] = None,
    ) -> CmdResult:
        """
        Execute a command on the remote host.
        Command will not be recorded.
        """
        if original_command is None:
            original_command = cmd
        ctx.debug("Execute: " + original_command)

        cmd_str = 'HISTIGNORE="*" ' + cmd

        try:
            await self.connect(ctx)
        except OSError:
            return CmdResult(
                original_command, 1, "SSH: Failed to connect to host.", ctx
            )

        process = await self._connection.run(
            cmd_str,
            term_type="xterm-color",
        )
        self.disconnect(ctx)

        if process.stdout is None:
            _stdout = ""
        elif isinstance(process.stdout, bytes):
            _stdout = process.stdout.decode()
        else:
            _stdout = process.stdout
        stdout = self._clean_log(_stdout)

        # try again as Windows
        if "'HISTIGNORE' is not recognized" in stdout:
            ctx.debug(
                "Potential Windows machine detected, retrying with updated settings."
            )
            self.is_windows_host = True
            await self.execute("Set-ExecutionPolicy RemoteSigned", ctx)
            return await self._execute_windows(cmd, ctx)

        if stdout:
            ctx.debug("stdout:\n" + stdout)
        else:
            ctx.debug("stdout:")

        returncode = 1
        if process.returncode is not None:
            returncode = process.returncode

        return CmdResult(original_command, returncode, stdout, ctx)

    async def _execute_sudo(
        self, cmd: str, ctx: HandlerContext, original_command: Optional[str] = None
    ) -> CmdResult:
        if original_command is None:
            original_command = cmd
        if self.sudo_requires_pass and self.password is None:
            return CmdResult(cmd, 1, "Sudo password is required, but not set!", ctx)

        if self.sudo_requires_pass:
            return await self._execute(
                f"echo -n {self.password} | sudo -S ls > /dev/null && " + cmd,
                ctx,
                cmd,
            )

        return await self.execute(cmd, ctx)

    async def _execute_windows(
        self,
        cmd: str,
        ctx: HandlerContext,
        original_command: Optional[str] = None,
    ) -> CmdResult:
        if original_command is None:
            original_command = cmd
        ctx.debug("Execute: " + original_command)

        cmd = cmd.replace('"', '\\"')
        if not cmd.startswith("powershell"):
            cmd_str = f'powershell -NoLogo -NoProfile "{cmd}"'
        else:
            cmd_str = cmd

        rcode_file = f"returncode-{ctx.normalized_task_id()}"
        output_file = f"output-{ctx.normalized_task_id()}"
        cmd_str = f"chcp 65001 & {cmd_str} > {output_file} & "
        cmd_str += f"echo %ERRORLEVEL% > {rcode_file}"

        try:
            await self.connect(ctx)
        except OSError:
            return CmdResult(
                original_command, 1, "SSH: Failed to connect to host.", ctx
            )

        await self._connection.run(
            cmd_str,
            term_type="xterm-color",
        )

        extra_args: dict[str, str] = {}
        if self.username is not None:
            extra_args["username"] = self.username

        if self.ssh_requires_pass:
            if self.password is not None:
                extra_args["password"] = self.password

        await asyncssh.scp(
            f"{self.host}:{rcode_file}",
            ctx.task_cache / rcode_file,
            **extra_args,  # type: ignore
        )
        with open(ctx.task_cache / rcode_file, encoding="utf-8") as f:
            returncode = int(f.read())
        os.remove(ctx.task_cache / rcode_file)

        await self._connection.run(
            f"del {rcode_file}",
            term_type="xterm-color",
        )

        await asyncssh.scp(
            f"{self.host}:{output_file}",
            ctx.task_cache / output_file,
            **extra_args,  # type: ignore
        )
        with open(ctx.task_cache / output_file, encoding="utf-8") as f:
            stdout = self._clean_log(f.read())
        os.remove(ctx.task_cache / output_file)

        await self._connection.run(
            f"del {output_file}",
            term_type="xterm-color",
        )

        if stdout:
            ctx.debug("stdout:\n" + stdout)
        else:
            ctx.debug("stdout:")

        self.disconnect(ctx)

        return CmdResult(original_command, returncode, stdout, ctx)

    def _clean_log(self, log: str) -> str:
        log = log.replace("\r\n", "\n")

        # strip all ansi characters
        ansi_escape = re.compile(
            r"""
                \x1B  # ESC
                (?:   # 7-bit C1 Fe (except CSI)
                    [@-Z\\-_]
                |     # or [ for CSI, followed by a control sequence
                    \[
                    [0-?]*  # Parameter bytes
                    [ -/]*  # Intermediate bytes
                    [@-~]   # Final byte
                )
            """,
            re.VERBOSE,
        )
        log = ansi_escape.sub("", log)

        sudo_log_str = "[sudo] password for "
        if self.username is not None:
            sudo_log_str += self.username
        else:
            sudo_log_str += os.getlogin()
        sudo_log_str += ": "
        log = log.replace(sudo_log_str, "")

        # weird windows stuff
        log = log.replace("0;Administrator: C:\\Windows\\system32\\conhost.exe", "")
        log = log.replace("0;C:\\Windows\\system32\\conhost.exe", "")
        log = log.replace("Active code page: 65001", "")

        # strip beeps, lol
        log = log.replace("\x07", "")

        log = log.removesuffix("\n")

        # Remove any extranious whitespace
        while "\n\n\n" in log:
            log = log.replace("\n\n\n", "\n\n")

        return log

    async def forward_port(
        self, ctx: HandlerContext, local_port: int, remote_port: int | None = None
    ) -> None:
        """
        Forwards a port using ssh.
        """
        if remote_port is None:
            remote_port = local_port

        forward_port = self._forwarded_ports.get(local_port)
        if forward_port is None:
            await self.connect(ctx)

            listener = await self._connection.forward_local_port(
                "127.0.0.1",
                local_port,
                "127.0.0.1",
                remote_port,
            )
            ctx.debug(
                f"SSH: Forwarding port '{local_port}' to '{self.host}:{remote_port}'."
            )

            self._forwarded_ports[local_port] = ForwardedPortListener(
                listener, local_port, remote_port
            )
        else:
            if forward_port.remote_port != remote_port:
                raise EikoDeployError(
                    "Trying to forward a port from a local port that is already in use."
                )

            ctx.debug(
                f"SSH: Reusing Forwarded port '{local_port}' to '{self.host}:{remote_port}'."
            )
            forward_port.connections += 1

    async def stop_forwarding_port(self, ctx: HandlerContext, local_port: int) -> None:
        """
        Stop forwarding a previously forwarded port.
        """
        forward_port = self._forwarded_ports.get(local_port)
        if forward_port is None or forward_port.connections == 0:
            ctx.warning("Tried to stop forwarding a port that is not being forwarded.")
            return

        self._forwarded_ports_stop_tasks.append(
            asyncio.create_task(
                self._forward_port_stop_delay(ctx, forward_port),
                name=f"forward_port_stop-{self.host}-{forward_port.connections}",
            )
        )

    async def _forward_port_stop_delay(
        self, ctx: HandlerContext, forward_port: ForwardedPortListener
    ) -> None:
        await asyncio.sleep(1)
        forward_port.connections -= 1
        if forward_port.connections == 0:
            forward_port.listener.close()
            await forward_port.listener.wait_closed()
            ctx.debug(
                f"SSH: stopped Forwarding port '{forward_port.local_port}' to '{self.host}:{forward_port.remote_port}'."
            )
            self.disconnect(ctx)


class HostHandler(Handler):
    """
    Represents a remote host.
    """

    __eiko_resource__ = "Host"

    async def execute(self, ctx: HandlerContext) -> None:
        if not isinstance(ctx.resource, HostModel):
            ctx.failed = True
            return

        if ctx.resource.password is None:
            if ctx.resource.ssh_requires_pass:
                ctx.error("ssh password is required, but not set!")
                ctx.failed = True
                return

        result = await ctx.resource.execute("echo $OSTYPE", ctx)
        if result.failed():
            return

        os_platform_promis = ctx.resource.raw_resource.promises["os_platform"]
        os_name_promise = ctx.resource.raw_resource.promises["os_name"]
        os_version_promis = ctx.resource.raw_resource.promises["os_version"]

        os_string = result.output.replace("\n", "")
        if os_string == "":
            os_platform_promis.set("windows", ctx)
            os_name_promise.set("windows", ctx)
            os_version_result = await ctx.resource.execute(
                "(Get-ComputerInfo).WindowsProductName",
                ctx,
            )
            os_version_promis.set(os_version_result.output.replace("\n", ""), ctx)
            await ctx.resource.execute("Set-ExecutionPolicy RemoteSigned", ctx)
            ctx.debug(f"OS Detection: {os_version_promis.resolve(str)}")
        elif os_string == "linux-gnu":
            os_platform_promis.set("linux-gnu", ctx)
            os_release = await ctx.resource.execute("cat /etc/os-release", ctx)
            os_info: dict[str, str] = {}
            for line in os_release.output.splitlines():
                key, value = line.split("=")
                os_info[key] = value

            os_name_promise.set(os_info["ID"], ctx)
            os_version_promis.set(os_info["VERSION_ID"], ctx)
            ctx.debug(
                "OS Detection: "
                f"{os_platform_promis.resolve(str)}-{os_name_promise.resolve(str)}-{os_version_promis.resolve(str)}"
            )

        else:
            os_platform_promis.set("unknown", ctx)
            os_name_promise.set("unknown", ctx)
            os_version_promis.set("unknown", ctx)

            ctx.debug(
                "OS Detection: "
                f"{os_platform_promis.resolve(str)}-{os_name_promise.resolve(str)}-{os_version_promis.resolve(str)}"
            )

        ctx.deployed = True

    async def cleanup(self, ctx: HandlerContext) -> None:
        ctx.debug("Cleaning up ssh connection.")
        if isinstance(ctx.resource, HostModel):
            await ctx.resource.wait_until_disconnected()


@eiko_plugin()
def get_pass(prompt: str = "Password: ") -> EikoProtectedStr:
    return EikoProtectedStr(getpass.getpass(prompt))


class CmdModel(EikoBaseModel):
    """
    A command that will be executed on a remote host.
    """

    __eiko_resource__ = "Cmd"

    host: HostModel
    cmd: str


class CmdHandler(Handler):
    """
    Represents a command executed on remote host.
    """

    __eiko_resource__ = "Cmd"

    async def execute(self, ctx: HandlerContext) -> None:
        if not isinstance(ctx.resource, CmdModel):
            ctx.failed = True
            return

        result = await ctx.resource.host.execute(ctx.resource.cmd, ctx)
        if result.failed():
            return

        ctx.deployed = True


class ScriptModel(EikoBaseModel):
    """
    A command that will be executed on a remote host.
    """

    __eiko_resource__ = "Script"

    host: HostModel
    script: str
    exec_shell: Optional[str] = None


class ScriptHandler(Handler):
    """
    Runs scripts on remote hosts.
    Writes given script to a file and executes said file.
    Then it cleans up the file.
    """

    __eiko_resource__ = "Script"

    async def execute(self, ctx: HandlerContext) -> None:
        if not isinstance(ctx.resource, ScriptModel):
            ctx.failed = True
            return

        if ctx.resource.exec_shell is None:
            platform = ctx.resource.host.os_platform.resolve(str)
            if platform == "linux-gnu":
                exec_shell = "bash"
            elif platform == "windows":
                exec_shell = "powershell"
            else:
                exec_shell = "sh"
        else:
            exec_shell = ctx.resource.exec_shell

        result = await ctx.resource.host.script(ctx.resource.script, exec_shell, ctx)
        if result.failed():
            return

        ctx.deployed = True


@eiko_plugin("hash")
def hash_plugin(string: str) -> str:
    return hex(hash(string))



# pylint: disable=unused-import
"""
Eikobot base types and helper methods,
usefull for module developers.
Imported from various places around the code.
"""
from . import human_readable, machine_readable
from .compiler.definitions.base_model import EikoBaseModel
from .compiler.definitions.base_types import (
    EikoBaseType,
    EikoBool,
    EikoDict,
    EikoFloat,
    EikoInt,
    EikoList,
    EikoPath,
    EikoPromise,
    EikoProtectedStr,
    EikoResource,
    EikoStr,
)
from .errors import EikoDeployError



"""
The file module helps to create and manipulate files.
"""
from pathlib import Path
from typing import Optional, Union

from jinja2 import Template

from eikobot.core.handlers import CRUDHandler, HandlerContext
from eikobot.core.helpers import EikoBaseModel
from eikobot.core.lib.std import HostModel
from eikobot.core.plugin import EikoPluginException, eiko_plugin


@eiko_plugin()
def read_file(path: Path) -> str:
    """
    Reads a file on the local machine.
    """
    try:
        _path = path.resolve(strict=True)
    except FileNotFoundError as e:
        raise EikoPluginException(f"Failed to read file '{path}'.") from e
    return _path.read_text(encoding="UTF-8")


@eiko_plugin()
def render_template(template: str, data: dict) -> str:
    """
    Takes a jinja template and data as input, and returns a rendered template.
    """
    jinja_template = Template(template)

    _data: dict[str, Union[str, int, bool, float, dict, list]] = {}
    for key, value in data.items():
        if not isinstance(key, str):
            raise EikoPluginException(
                "render_template requires keys in its data dict to be strings."
            )

        if not isinstance(value, (str, int, bool, float, dict, list)):
            raise EikoPluginException(
                "render_template requires values to be strings, "
                "bools, floats, integers or dicts and lists."
            )

        _data[key] = value

    return jinja_template.render(_data)


class FileModel(EikoBaseModel):
    """
    Represents the eiko 'File' class
    """

    __eiko_resource__ = "File"

    host: HostModel
    path: Path
    content: str
    owner: Optional[str] = None
    group: Optional[str] = None
    mode: str = "664"
    requires_sudo: bool = False


@eiko_plugin()
def parse_rwx_mode(mode: str) -> str:
    """Parses a mode in rw- style and turns it in to an int style."""
    conversion_table = {
        "x": 1,
        "w": 2,
        "r": 4,
    }
    if len(mode) == 10:
        mode = mode[1:]

    if len(mode) != 9:
        raise EikoPluginException

    oct_1 = 0
    for char in mode[0:2]:
        oct_1 += conversion_table.get(char, 0)
    n_mode = str(oct_1)

    oct_2 = 0
    for char in mode[3:5]:
        oct_2 += conversion_table.get(char, 0)
    n_mode += str(oct_2)

    oct_3 = 0
    for char in mode[6:]:
        oct_3 += conversion_table.get(char, 0)
    n_mode += str(oct_3)

    return n_mode


class FileHandler(CRUDHandler):
    """Deploys a file on the target machine using ssh."""

    __eiko_resource__ = "File"

    async def create(self, ctx: HandlerContext) -> None:
        if not isinstance(ctx.resource, FileModel):
            ctx.failed = True
            return

        if ctx.resource.host.is_windows_host:
            await self._create_win(ctx, ctx.resource)

        else:
            await self._create(ctx, ctx.resource)

    async def _create(self, ctx: HandlerContext, resource: FileModel) -> None:
        if resource.requires_sudo:
            sudo = "sudo "
        else:
            sudo = ""

        result = await resource.host.execute(
            f"{sudo}mkdir -p {resource.path.parent}", ctx
        )
        if result.returncode != 0:
            ctx.failed = True
            return

        result = await resource.host.execute(
            f'echo -n "{resource.content}" | {sudo}tee {resource.path}',
            ctx,
        )
        if result.returncode != 0:
            ctx.failed = True
            return

        result = await resource.host.execute(
            f"{sudo}chmod {resource.mode} " f"{resource.path}",
            ctx,
        )
        if result.returncode != 0:
            ctx.failed = True
            return

        if resource.owner is not None:
            result = await resource.host.execute(
                f"{sudo}chown {resource.owner} {resource.path}",
                ctx,
            )
            if result.returncode != 0:
                ctx.failed = True
                return

        if resource.group is not None:
            result = await resource.host.execute(
                f"{sudo}chgrp {resource.group} {resource.path}",
                ctx,
            )
            if result.returncode != 0:
                ctx.failed = True
                return

        ctx.deployed = True

    async def _create_win(self, ctx: HandlerContext, resource: FileModel) -> None:
        result = await resource.host.execute(f"mkdir {resource.path.parent}", ctx)
        if result.returncode != 0:
            ctx.failed = True
            return

        with open(ctx.task_cache / resource.path.name, "w", encoding="utf-8") as f:
            f.write(resource.content)

        await resource.host.scp_to(
            Path(ctx.task_cache / resource.path.name),
            ctx,
            str(resource.path),
        )

        ctx.deployed = True

    async def read(self, ctx: HandlerContext) -> None:
        if not isinstance(ctx.resource, FileModel):
            ctx.failed = True
            return

        if ctx.resource.host.is_windows_host:
            await self._read_win(ctx, ctx.resource)

        else:
            await self._read(ctx, ctx.resource)

    async def _read(self, ctx: HandlerContext, resource: FileModel) -> None:
        if resource.requires_sudo:
            sudo = "sudo "
        else:
            sudo = ""

        cat_result = await resource.host.execute(f"{sudo}cat {resource.path}", ctx)
        if cat_result.returncode == 0:
            ctx.deployed = True
            if cat_result.output != resource.content:
                ctx.add_change("content", cat_result.output)

            ls_result = await resource.host.execute(f"{sudo}ls -l {resource.path}", ctx)
            if ls_result.returncode == 0:
                if isinstance(ls_result.output, str):
                    ls_parsed = ls_result.output.split(" ")
                    if parse_rwx_mode(ls_parsed[0]) != resource.mode:
                        ctx.add_change("mode", parse_rwx_mode(ls_parsed[0]))
                    if resource.owner is not None and ls_parsed[2] != resource.owner:
                        ctx.add_change("owner", ls_parsed[2])
                    if resource.group is not None and ls_parsed[3] != resource.group:
                        ctx.add_change("group", ls_parsed[3])
                else:
                    ctx.failed = True

    async def _read_win(self, ctx: HandlerContext, resource: FileModel) -> None:
        cat_result = await resource.host.execute(f"cat {resource.path}", ctx)
        if cat_result.returncode == 0:
            ctx.deployed = True
            if cat_result.output != resource.content:
                ctx.add_change("content", cat_result.output)

    async def update(self, ctx: HandlerContext) -> None:
        if not isinstance(ctx.resource, FileModel):
            ctx.failed = True
            return

        if ctx.resource.host.is_windows_host:
            await self._create_win(ctx, ctx.resource)

        else:
            await self._update(ctx, ctx.resource)

    async def _update(self, ctx: HandlerContext, resource: FileModel) -> None:
        if resource.requires_sudo:
            sudo = "sudo "
        else:
            sudo = ""

        if ctx.changes.get("content") is not None:
            result = await resource.host.execute(
                f'echo -n "{resource.content}" | {sudo}tee {resource.path}',
                ctx,
            )
            if result.failed():
                return

        if ctx.changes.get("mode") is not None:
            result = await resource.host.execute(
                f"{sudo}chmod {resource.mode} " f"{resource.path}",
                ctx,
            )
            if result.failed():
                return

        if ctx.changes.get("owner") is not None:
            result = await resource.host.execute(
                f"{sudo}chown {resource.owner} {resource.path}",
                ctx,
            )
            if result.failed():
                return

        if ctx.changes.get("group") is not None:
            result = await resource.host.execute(
                f"{sudo}chgrp {resource.group} {resource.path}",
                ctx,
            )
            if result.failed():
                return

        ctx.deployed = True



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-statements
from pathlib import Path

import pytest

from eikobot.core.compiler import Compiler
from eikobot.core.compiler._parser import Parser
from eikobot.core.compiler.definitions._resource import EikoResourceDefinition
from eikobot.core.compiler.definitions.base_types import (
    EikoBool,
    EikoEnumValue,
    EikoInt,
    EikoList,
    EikoNone,
    EikoResource,
    EikoStr,
)
from eikobot.core.compiler.definitions.context import CompilerContext
from eikobot.core.compiler.definitions.typedef import EikoTypeDef
from eikobot.core.errors import EikoCompilationError


def test_basic_ops(eiko_basic_ops_file: Path) -> None:
    parser = Parser(eiko_basic_ops_file)
    ast_list = list(parser.parse())
    context = CompilerContext("__test__", {})

    result_1 = ast_list[0].compile(context)
    assert isinstance(result_1, EikoInt)
    assert result_1.value == 6

    result_2 = ast_list[1].compile(context)
    assert isinstance(result_2, EikoInt)
    assert result_2.value == -14

    result_3 = ast_list[2].compile(context)
    assert isinstance(result_3, EikoInt)
    assert result_3.value == -13

    result_4 = ast_list[3].compile(context)
    assert isinstance(result_4, EikoStr)
    assert result_4.value == "string 1 + string 2"

    result_5 = ast_list[4].compile(context)
    assert isinstance(result_5, EikoInt)
    assert result_5.value == -4

    result_6 = ast_list[5].compile(context)
    assert isinstance(result_6, EikoStr)
    assert result_6.value == "hahaha"

    result_7 = ast_list[6].compile(context)
    assert isinstance(result_7, EikoStr)
    assert result_7.value == "auto concat string"


def test_basic_compiler(eiko_file_3: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_file_3)
    test_1 = compiler.context.get("test_1")
    assert isinstance(test_1, EikoResource)

    _ip = test_1.properties["ip"]
    assert isinstance(_ip, EikoStr)
    assert _ip.value == "192.168.0.1"

    ip_2 = test_1.properties["ip_2"]
    assert isinstance(ip_2, EikoStr)
    assert ip_2.value == "192.168.1.1"

    ip_1 = compiler.context.get("ip_1")
    assert isinstance(ip_1, EikoStr)
    assert ip_1.value == "192.168.0.1"


def test_nested_properties(nested_properties_file: Path) -> None:
    compiler = Compiler()
    compiler.compile(nested_properties_file)
    var_b = compiler.context.get("b")
    assert isinstance(var_b, EikoStr)


def test_typedef(eiko_typedef: Path) -> None:
    compiler = Compiler()
    with pytest.raises(EikoCompilationError):
        compiler.compile(eiko_typedef)

    string_alias = compiler.context.get("string_alias")
    assert isinstance(string_alias, EikoTypeDef)

    str_1 = compiler.context.get("str_1")
    assert isinstance(str_1, EikoStr)
    assert str_1.type.name == "string_alias"

    str_2 = compiler.context.get("str_2")
    assert isinstance(str_2, EikoStr)
    assert str_2.type.name == "string_alias"

    ipv4address = compiler.context.get("IPv4Address")
    assert isinstance(ipv4address, EikoTypeDef)

    var_a = compiler.context.get("a")
    assert isinstance(var_a, EikoStr)
    assert var_a.value == "10.0.0.0"

    res_1 = compiler.context.get("res_1")
    assert isinstance(res_1, EikoResource)

    prop_1 = res_1.get("prop_1")
    assert isinstance(prop_1, EikoStr)
    assert prop_1.value == "192.168.0.1"

    net_port = compiler.context.get("net_port")
    assert isinstance(net_port, EikoInt)
    assert net_port.value == 80
    assert net_port.type.name == "WellKnownPort"
    assert net_port.type.super is not None
    assert net_port.type.super.name == "NetworkPort"


def test_resource_compilation(eiko_file_1: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_file_1)

    var_test_1 = compiler.context.get("test_1")
    assert isinstance(var_test_1, EikoResource)

    prop_ip = var_test_1.get("ip")
    assert isinstance(prop_ip, EikoStr)
    assert prop_ip.value == "192.168.0.1"

    prop_ip_2 = var_test_1.get("ip_2")
    assert isinstance(prop_ip_2, EikoStr)
    assert prop_ip_2.value == "192.168.1.1"


def test_unions(eiko_union_file: Path) -> None:
    compiler = Compiler()
    with pytest.raises(EikoCompilationError):
        compiler.compile(eiko_union_file)

    var_a = compiler.context.get("a")
    assert isinstance(var_a, EikoStr)

    var_b = compiler.context.get("b")
    assert isinstance(var_b, EikoNone)


def test_forward_declare(tmp_eiko_file: Path) -> None:
    compiler = Compiler()
    with open(tmp_eiko_file, "w", encoding="utf-8") as f:
        f.write("a: str\n" + 'a = "test_string"')

    compiler.compile(tmp_eiko_file)

    var_a = compiler.context.get("a")
    assert isinstance(var_a, EikoStr)
    assert var_a.value == "test_string"


def test_custom_constructor(eiko_constructor_file: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_constructor_file)

    res_def = compiler.context.get("ConstructorTestResource")
    assert isinstance(res_def, EikoResourceDefinition)

    var_a = compiler.context.get("a")
    assert isinstance(var_a, EikoResource)

    prop_1 = var_a.get("prop_1")
    assert isinstance(prop_1, EikoStr)
    assert prop_1.value == "test"

    prop_2 = var_a.get("prop_2")
    assert isinstance(prop_2, EikoInt)
    assert prop_2.value == 1

    prop_3 = var_a.get("prop_3")
    assert isinstance(prop_3, EikoStr)
    assert prop_3.value == "testtest"


def test_inheritance(eiko_inheritance_file: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_inheritance_file)

    var_a = compiler.context.get("a")
    assert isinstance(var_a, EikoResource)
    assert var_a.type.name == "BaseRes"
    assert var_a.to_py() == {
        "prop_1": "a",
        "prop_2": 1,
    }

    var_b = compiler.context.get("b")
    assert isinstance(var_b, EikoResource)
    assert var_b.type.name == "SubRes"
    assert var_b.type.super is not None
    assert var_b.type.super.name == "BaseRes"
    assert var_b.to_py() == {
        "prop_1": "a",
        "prop_2": 1,
        "prop_3": 1,
    }

    var_c = compiler.context.get("c")
    assert isinstance(var_c, EikoResource)
    assert var_c.type.name == "SubSubRes"
    assert var_c.type.super is not None
    assert var_c.type.super.name == "SubRes"
    assert var_c.type.super.super is not None
    assert var_c.type.super.super.name == "BaseRes"
    assert var_c.to_py() == {
        "prop_1": "a",
        "prop_2": 1,
        "prop_3": "a",
        "prop_4": 1,
    }

    var_d = compiler.context.get("d")
    assert isinstance(var_d, EikoResource)
    assert var_d.type.name == "SubResPropOverwite"
    assert var_d.type.super is not None
    assert var_d.type.super.name == "SubRes"
    assert var_d.type.super.super is not None
    assert var_d.type.super.super.name == "BaseRes"
    assert var_d.to_py() == {
        "prop_1": "a",
        "prop_2": 1,
        "prop_3": "a",
    }

    var_e = compiler.context.get("e")
    assert isinstance(var_e, EikoResource)
    assert var_e.type.name == "Test"
    assert var_e.type.super is not None
    assert var_e.type.super.name == "Object"
    assert var_e.to_py() == {
        "prop_1": var_a.to_py(),
    }

    var_f = compiler.context.get("f")
    assert isinstance(var_f, EikoResource)
    assert var_f.type.name == "Test"
    assert var_f.type.super is not None
    assert var_f.type.super.name == "Object"
    assert var_f.to_py() == {
        "prop_1": var_b.to_py(),
    }

    var_g = compiler.context.get("g")
    assert isinstance(var_g, EikoResource)
    assert var_g.type.name == "Test"
    assert var_g.type.super is not None
    assert var_g.type.super.name == "Object"
    assert var_g.to_py() == {
        "prop_1": var_c.to_py(),
    }

    var_h = compiler.context.get("h")
    assert isinstance(var_h, EikoResource)
    assert var_h.type.name == "Test"
    assert var_h.type.super is not None
    assert var_h.type.super.name == "Object"
    assert var_h.to_py() == {
        "prop_1": var_d.to_py(),
    }

    var_i = compiler.context.get("i")
    assert isinstance(var_i, EikoResource)
    assert var_i.type.name == "TripleDotInherit"
    assert var_i.to_py() == {
        "prop_1": "a",
        "prop_2": 1,
    }


def test_enum(eiko_enum_file: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_enum_file)

    var_test_1 = compiler.context.get("test_1")
    assert isinstance(var_test_1, EikoResource)
    assert var_test_1.type.name == "TestRes"
    var_test_1_enum_opt_2 = var_test_1.properties.get("prop_1")
    assert isinstance(var_test_1_enum_opt_2, EikoEnumValue)
    assert var_test_1_enum_opt_2.value == "option_2"

    var_test_2 = compiler.context.get("test_2")
    assert isinstance(var_test_2, EikoResource)
    assert var_test_2.type.name == "TestRes"
    var_test_2_enum_opt_1 = var_test_2.properties.get("prop_1")
    assert isinstance(var_test_2_enum_opt_1, EikoEnumValue)
    assert var_test_2_enum_opt_1.value == "option_1"

    var_enum_to_str = compiler.context.get("enum_to_str")
    assert isinstance(var_enum_to_str, EikoStr)
    assert var_enum_to_str.value == "option_3"


@pytest.mark.parametrize(
    "input_str,outcome",
    [
        ('a = 1\nb = type(a) == "int"', True),
        ('a = 1\nb = type(a) == "str"', False),
        ('a = "1"\nb = type(a) == "str"', True),
        ('a = True\nb = type(a) == "bool"', True),
        ('a = [True]\nb = type(a) == "list[bool]"', True),
    ],
)
def test_type_plugin(tmp_eiko_file: Path, input_str: str, outcome: bool) -> None:
    compiler = Compiler()

    with open(tmp_eiko_file, "w", encoding="utf-8") as f:
        f.write(input_str)

    compiler.compile(tmp_eiko_file)
    b = compiler.context.get("b")
    assert isinstance(b, EikoBool)
    assert b.value == outcome


def test_for(eiko_for_file: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_for_file)

    results_list = compiler.context.get("test_list")
    assert isinstance(results_list, EikoList)

    assert isinstance(results_list.elements[0], EikoStr)
    assert results_list.elements[0].value == "hello"

    assert isinstance(results_list.elements[1], EikoStr)
    assert results_list.elements[1].value == "haha"

    assert isinstance(results_list.elements[2], EikoInt)
    assert results_list.elements[2].value == 12

    assert isinstance(results_list.elements[3], EikoStr)
    assert results_list.elements[3].value == "key_1"

    assert isinstance(results_list.elements[4], EikoInt)
    assert results_list.elements[4].value == 1


def test_membership(eiko_membership_file: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_membership_file)

    a = compiler.context.get("a")
    assert isinstance(a, EikoBool) and a.value

    b = compiler.context.get("b")
    assert isinstance(b, EikoBool) and not b.value

    c = compiler.context.get("c")
    assert isinstance(c, EikoBool) and c.value

    d = compiler.context.get("d")
    assert isinstance(d, EikoBool) and not d.value

    e = compiler.context.get("e")
    assert isinstance(e, EikoBool) and e.value

    f = compiler.context.get("f")
    assert isinstance(f, EikoBool) and not f.value

    g = compiler.context.get("g")
    assert isinstance(g, EikoBool) and g.value



"""
For help with creating, distributing and installing packages.
"""
import asyncio
import copy
import json
import os
import shutil
import subprocess
import sys
import tarfile
import tomllib
from dataclasses import dataclass
from pathlib import Path
from urllib.parse import urlparse

import aiohttp
from packaging import version
from pydantic import BaseModel, ValidationError

from .. import human_readable, logger
from ..compiler.importlib import INTERNAL_LIB_PATH
from ..errors import EikoError

CACHE_PATH = Path(__file__).parent / "cache"
CACHE_PATH.mkdir(exist_ok=True)
LIB_PATH = Path(__file__).parent / "lib"
LIB_PATH.mkdir(exist_ok=True)

PKG_INDEX_PATH = Path(__file__).parent / "index.json"

GH_HEADERS = {
    "X-GitHub-Api-Version": "2022-11-28",
    "Accept": "application/vnd.github+json",
}
GH_TOKEN = os.environ.get("GITHUB_TOKEN", None)
if GH_TOKEN is not None:
    GH_HEADERS["Authorization"] = f"Bearer {GH_TOKEN}"


class EikoPackageError(EikoError):
    """An error that occured during the eiko compilation process."""

    def __init__(self, reason: str, *args: object) -> None:
        super().__init__(reason, *args)


class PackageData(BaseModel):
    """
    All the settings one might
    find in an eiko.toml file.
    """

    name: str
    source_dir: Path
    version: str | None = None
    description: str | None = None
    long_description: str | None = None
    long_description_content_type: str | None = None
    author: str | None = None
    author_email: str | None = None
    license: str | None = None
    eikobot_version: str | None = None
    eikobot_requires: list[str] = []
    python_requires: list[str] = []
    github_name: str | None = None

    def pkg_name_version(self) -> str:
        """
        Returns the name that was used for the install dir
        under package_manager/lib.
        """
        name = self.name
        if self.version is not None:
            name += "-" + self.version

        return name


def get_pkg_index() -> dict[str, "PackageData"]:
    """
    Constructs the package index and returns it
    """
    pkg_index: dict[str, "PackageData"] = {}
    for pkg_dir in LIB_PATH.glob("*"):
        pkg_data = read_pkg_toml(pkg_dir / "eiko.toml")
        pkg_index[pkg_data.name] = pkg_data

    return pkg_index


def read_pkg_toml(path: Path) -> PackageData:
    """
    Reads a toml and returns it's settings as python object.
    """
    logger.debug("Reading eiko toml file.")
    toml = tomllib.loads(path.read_text(encoding="utf-8"))
    pkg_toml = toml.get("eiko", {}).get("package")
    if pkg_toml is None:
        raise EikoPackageError("eiko.toml does not contain an 'eiko.package' section.")
    try:
        pkg_data = PackageData(**pkg_toml)
    except ValidationError as e:
        raise EikoPackageError(f"Failed to parse eiko.toml: {e}") from e

    return pkg_data


def build_pkg(pkg_data: PackageData | None = None) -> Path:
    """Builds a package in the cwd"""
    logger.info("Building package.")
    eiko_toml_path = Path("eiko.toml")
    if pkg_data is None:
        if not eiko_toml_path.exists():
            raise EikoPackageError("eiko.toml file missing.")
        pkg_data = read_pkg_toml(eiko_toml_path)

    dist = Path("dist")
    dist.mkdir(exist_ok=True)
    if not pkg_data.source_dir.exists():
        raise EikoPackageError(f"Not such source directory: '{pkg_data.source_dir}'.")

    if not pkg_data.source_dir.is_dir():
        raise EikoPackageError(f"'{pkg_data.source_dir}' is not a directory.")

    logger.debug("Setting up build env.")
    build_dir = dist / "build"
    shutil.rmtree(build_dir, ignore_errors=True)
    build_dir.mkdir()
    # Compile every file here to make sure they are valid?
    # This would also be the place to check for dangerous sources.
    shutil.copytree(
        pkg_data.source_dir,
        build_dir / pkg_data.source_dir,
        ignore=shutil.ignore_patterns("__pycache__"),
    )
    shutil.copy(eiko_toml_path, build_dir)

    logger.debug("Creating tar archive.")
    dist_name = pkg_data.pkg_name_version()
    dist_file_name = dist_name + ".eiko.tar.gz"
    dist_file = dist / dist_file_name

    with tarfile.open(dist_file, "w:gz") as archive:
        archive.add(build_dir, arcname=dist_name)

    logger.debug("cleaning up build env.")
    shutil.rmtree(build_dir, ignore_errors=True)

    logger.info(f"Build package '{dist_name}'")
    return dist_file


async def install_pkg(pkg_def: str) -> None:
    """
    This functions gets a string prompt and will try to figure out
    if the package is http(s), git or a path.
    """
    async with aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(10)) as session:
        if pkg_def.startswith("http://") or pkg_def.startswith("https://"):
            await _download_pkg(pkg_def, session)

        elif pkg_def.startswith("GH://"):
            await _download_pkg_gh(pkg_def, session)

        elif pkg_def.endswith(".eiko.tar.gz"):
            await _install_pkg_from_path(Path(pkg_def))

        else:
            raise EikoPackageError("Failed to properly parse package url or path.")


async def _download_pkg(url: str, session: aiohttp.ClientSession) -> None:
    if url.endswith(".eiko.tar.gz"):
        await _download_to_cache(url, session)

    else:
        raise EikoPackageError(
            "Currently only direct links to an eiko package are supported."
        )


def _get_gh_package(github_name: str) -> PackageData | None:
    pkg_index = get_pkg_index()
    for package in pkg_index.values():
        if (
            package.github_name is not None
            and package.github_name.lower() == github_name.lower()
        ):
            return package

    return None


@dataclass
class GHAsset:
    name: str
    url: str


async def _download_pkg_gh(pkg_def: str, session: aiohttp.ClientSession) -> None:
    """
    Download a package straight from github, without extra fuzz.
    """
    logger.debug(f"[{pkg_def}] Parsing github link.")
    _pkg_def = pkg_def.replace("GH://", "")
    _version: version.Version | None = None
    if "==" in _pkg_def:
        _pkg_def, _version = _parse_pkg_version("GH://", _pkg_def)

    prev_pkg = _get_gh_package(_pkg_def)
    if prev_pkg is not None:
        if _version is None or (
            prev_pkg.version is not None and version.parse(prev_pkg.version) == _version
        ):
            logger.info(f"Package '{_pkg_def}' already installed.")
            return

        _uninstall_pkg(prev_pkg)

    assets = await _get_gh_assets(
        f"https://api.github.com/repos/{_pkg_def}", pkg_def, session
    )

    if len(assets) == 0:
        raise EikoPackageError(f"Github package '{pkg_def}' has no releases or assets.")

    pkg: GHAsset | None = None
    _, pkg_name = _pkg_def.split("/")
    if _version is None:
        pkg = assets[0]

    else:
        for asset in assets:
            if pkg_name.lower() in asset.name.lower() and str(_version) in asset.name:
                pkg = asset
                break

    if pkg is None:
        raise EikoPackageError(f"No package that matches '{pkg_def}' was found.")

    await _download_pkg(pkg.url, session)


def _parse_pkg_version(protocol: str, pkg_def: str) -> tuple[str, version.Version]:
    try:
        _pkg_def, unparsed_version = pkg_def.split("==")
        return _pkg_def, version.parse(unparsed_version)
    except version.InvalidVersion as e:
        raise EikoPackageError(
            f"failed to parse package version `{unparsed_version}`."
        ) from e
    except ValueError as e:
        raise EikoPackageError(f"Failed to parse '{protocol}{pkg_def}'.") from e


async def _get_gh_assets(
    gh_url: str,
    pkg_def: str,
    session: aiohttp.ClientSession,
) -> list[GHAsset]:
    logger.debug(f"[{pkg_def}] Getting github releases.")
    releases_resp = await session.get(gh_url + "/releases", headers=GH_HEADERS)
    if 200 < releases_resp.status > 299:
        await session.close()
        raise EikoPackageError(
            f"Repo not found on github: '{pkg_def}'. (HTTP {releases_resp.status})"
        )

    releases = await releases_resp.json()
    if len(releases) == 0:
        return []

    logger.debug(f"[{pkg_def}] Getting github assets.")
    assets: list[GHAsset] = []
    for release in releases:
        assets_resp = await session.get(release["assets_url"])
        unparsed_assets = await assets_resp.json()
        for _asset in unparsed_assets:
            assets.append(GHAsset(_asset["name"], _asset["browser_download_url"]))

    return assets


async def _download_to_cache(url: str, session: aiohttp.ClientSession) -> None:
    pkg_path = CACHE_PATH / Path(urlparse(url).path).name
    if not pkg_path.exists():
        try:
            response = await session.get(url)
            dl_size = int(response.headers.get("CONTENT-LENGTH", 0))
            logger.info(f"Fetching '{url}' [{human_readable(dl_size)}]")

            with open(pkg_path, "wb") as f:
                async for data, _ in response.content.iter_chunks():
                    f.write(data)

            logger.debug(f"Downloaded {human_readable(dl_size)}.")

        except aiohttp.ClientError as e:
            raise EikoPackageError(f"Failed to download {url}: {e}") from e
    else:
        logger.debug(f"Using cached version of '{pkg_path.name}'.")

    await _install_pkg_from_cache(pkg_path.name)


async def _install_pkg_from_path(pkg_path: Path) -> None:
    """
    Install a package using a path to an .eiko.tar.gz file.
    """
    if not pkg_path.exists():
        raise EikoPackageError(f"Package path does not exist: '{pkg_path}'.")
    logger.debug("Adding archive to cache.")
    shutil.copy(pkg_path, CACHE_PATH)
    await _install_pkg_from_cache(pkg_path.name)


async def _install_pkg_from_cache(archive_name: str) -> None:
    pkg_name = archive_name.removesuffix(".eiko.tar.gz")
    pkg_index = get_pkg_index()
    if (LIB_PATH / archive_name).exists():
        logger.debug("Archive is already unpacked.")

    else:
        logger.debug("Unpacking archive.")
        with tarfile.open(CACHE_PATH / archive_name, "r:gz") as archive:
            archive.extractall(LIB_PATH)

    pkg_lib_path = LIB_PATH / pkg_name
    pkg_data = read_pkg_toml(pkg_lib_path / "eiko.toml")
    prev_pkg = pkg_index.get(pkg_data.name)
    if prev_pkg is not None:
        if pkg_data.version != prev_pkg.version or pkg_data.version is None:
            _uninstall_pkg(prev_pkg)
        else:
            logger.info(
                f"Package '{pkg_data.name}=={pkg_data.version}' already installed."
            )
            return

    if pkg_data.version is None:
        logger.debug(f"Installing '{pkg_data.name}'.")
    else:
        logger.debug(f"Installing '{pkg_data.name}=={pkg_data.version}'.")

    install_py_deps(pkg_data.python_requires)

    logger.debug(f"Installing requirements for '{pkg_name}'.")
    await install_pkgs(pkg_data.eikobot_requires)

    try:
        shutil.copytree(
            pkg_lib_path / pkg_data.source_dir,
            INTERNAL_LIB_PATH / pkg_data.source_dir,
            dirs_exist_ok=False,
        )
    except FileExistsError:
        # This is a bug in the package index
        # I have no idea what causes it, but this is a hack around the issue.
        os.remove(INTERNAL_LIB_PATH / pkg_data.source_dir)
        os.symlink(
            pkg_lib_path / pkg_data.source_dir, INTERNAL_LIB_PATH / pkg_data.source_dir
        )

    if pkg_data.version is None:
        logger.info(f"Installed '{pkg_data.name}'.")
    else:
        logger.info(f"Installed '{pkg_data.name}=={pkg_data.version}'.")


def install_py_deps(deps: list[str]) -> None:
    """
    Install Python dependencies using pip running in a subprocess.
    """
    logger.debug("Installing python requirements.")
    if logger.LOG_LEVEL == logger.LogLevel.DEBUG:
        proc = subprocess.run(
            [sys.executable, "-m", "pip", "install", *deps],
            check=False,
        )
    else:
        proc = subprocess.run(
            [sys.executable, "-m", "pip", "install", *deps],
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    if proc.returncode != 0:
        logger.error(f"'pip install {' '.join(deps)}' failed.")
        if logger.LOG_LEVEL != logger.LogLevel.DEBUG:
            print(proc.stderr.decode())

        sys.exit(1)


def uninstall_pkg(name: str) -> None:
    """
    Uninstalls a package.
    """
    pkg_data = get_pkg_index().get(name)
    if pkg_data is None:
        logger.error(f"Package not found: '{name}'")
        return

    _uninstall_pkg(pkg_data)


def _uninstall_pkg(pkg_data: PackageData) -> None:
    logger.debug(f"Uninstalling '{pkg_data.pkg_name_version()}'")
    try:
        shutil.rmtree(INTERNAL_LIB_PATH / pkg_data.source_dir)
    except FileNotFoundError:
        pass

    shutil.rmtree(LIB_PATH / pkg_data.pkg_name_version())
    logger.info(f"Uninstalled '{pkg_data.pkg_name_version()}'")


async def install_pkgs(pkg_list: list[str]) -> None:
    for pkg in pkg_list:
        await install_pkg(pkg)


async def create_github_release() -> None:
    """
    Does the follozing in order:
    - Checks to make sure the version doesn't exist on Github
    - Builds the package
    - Tags it on the git tree and pushes the tags
    - Creates a release on Github
    - Uploads the package as a release asset
    """
    if GH_TOKEN is None:
        raise EikoPackageError("Environment variable 'GITHUB_TOKEN' must be set.")

    logger.info("Releasing to github.")
    eiko_toml_path = Path("eiko.toml")
    if not eiko_toml_path.exists():
        raise EikoPackageError("eiko.toml file missing.")
    pkg_data = read_pkg_toml(eiko_toml_path)

    if pkg_data.github_name is None:
        raise EikoPackageError("'github_name' must be set in eiko.toml.")
    owner, repo = pkg_data.github_name.split("/")

    if pkg_data.version is None:
        raise EikoPackageError("'version' must be set in eiko.toml.")

    git_status = await asyncio.subprocess.create_subprocess_shell(
        "git status --porcelain",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, _ = await git_status.communicate()
    if stdout.decode():
        raise EikoPackageError("Git tree is not clean!\n" + stdout.decode())

    logger.debug("Checking github repo for existing packages.")
    session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(10))
    gh_assets = await _get_gh_assets(
        f"https://api.github.com/repos/{pkg_data.github_name}",
        pkg_data.github_name,
        session,
    )

    for asset in gh_assets:
        if pkg_data.version in asset.name:
            await session.close()
            raise EikoPackageError(
                f"A release with version number '{pkg_data.version}' "
                "already exists on Github."
            )

    archive = build_pkg(pkg_data)
    logger.info("Creating git tags.")
    await _git_tag_and_push(pkg_data)

    logger.info("Creating release on github and uploading data.")
    release_url = await _create_gh_release(owner, repo, pkg_data, session)
    await _upload_gh_asset(archive, release_url, session)

    await session.close()
    logger.info("Successfully created release!")


async def _git_tag_and_push(pkg_data: PackageData) -> None:
    await asyncio.subprocess.create_subprocess_shell(
        f"git tag {pkg_data.version}",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    await asyncio.subprocess.create_subprocess_shell(
        f"git tag v{pkg_data.version}",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    await asyncio.subprocess.create_subprocess_shell(
        "git push --tags",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )


async def _create_gh_release(
    owner: str,
    repo: str,
    pkg_data: PackageData,
    session: aiohttp.ClientSession,
) -> str:
    url = f"https://api.github.com/repos/{owner}/{repo}/releases"
    data = {
        "owner": owner,
        "repo": repo,
        "tag_name": pkg_data.version,
        "name": f"v{pkg_data.version}",
        "draft": False,
        "prerelease": False,
        "generate_release_notes": False,
    }
    headers = {
        **GH_HEADERS,
        "content-type": "application/json",
    }
    response = await session.post(
        url,
        data=json.dumps(data),
        headers=headers,
    )
    if 200 < response.status > 299:
        await session.close()
        raise EikoPackageError(
            "Failed to create a release on Github. "
            f"(HTTP {response.status}: {await response.text()})"
        )

    return (await response.json())["url"]  # type: ignore


async def _upload_gh_asset(
    asset: Path, release_url: str, session: aiohttp.ClientSession
) -> None:
    url = release_url.replace("api.github.com", "uploads.github.com")
    url += f"/assets?name={asset.name}"
    response = await session.post(url, data=asset.read_bytes(), headers=GH_HEADERS)
    if 200 < response.status > 299:
        await session.close()
        raise EikoPackageError(
            "Failed to upload asset to Github release. "
            f"(HTTP {response.status}: {await response.text()})"
        )



"""
The importlib allows for importing
of both Eiko and Python code.
"""
import importlib.util
from dataclasses import dataclass
from inspect import getfullargspec, getmembers, isclass, isfunction
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Callable, Optional

from .. import logger
from ..errors import EikoCompilationError
from ..handlers import CRUDHandler, Handler
from ._token import Token
from .definitions.base_model import EikoBaseModel
from .definitions.function import PluginArg, PluginDefinition

if TYPE_CHECKING:
    from .definitions.context import CompilerContext

INTERNAL_LIB_PATH = Path(__file__).parent.parent.resolve() / "lib"
PATHS: list[Path] = [INTERNAL_LIB_PATH, Path(".")]


@dataclass
class Module:
    """A resolved module that can be imported by being compiled."""

    name: str
    path: Path
    context: "CompilerContext"

    def __post_init__(self) -> None:
        self.submodules: list["Module"] = []
        self.context.set_path(self.path)


def resolve_import(
    import_path: list[str], context: "CompilerContext"
) -> Optional[Module]:
    """
    Tries to import a given path.
    """
    for _path in PATHS:
        result = _resolve_import(import_path.copy(), _path, context)
        if result is not None:
            return result

    return None


def _resolve_import(
    import_path: list[str], parent: Path, context: "CompilerContext"
) -> Optional[Module]:
    current = import_path[0]
    import_path.remove(current)

    current_dir = parent / current
    if current_dir.exists() and current_dir.is_dir():
        init_file = current_dir / "__init__.eiko"
        if init_file.exists():
            if len(import_path) == 0:
                new_context = context.get_or_set_context(current)
                module = Module(current_dir.stem, init_file, new_context)
                _get_submodules(module)
                return module

            new_context = context.get_or_set_context(current)
            return _resolve_import(import_path, current_dir, new_context)

        return None

    if len(import_path) == 0:
        file_path = parent / (current + ".eiko")
        if file_path.exists():
            new_context = context.get_or_set_context(current)
            return Module(file_path.stem, file_path, new_context)

    return None


def test_path_is_module(path: Path, token: Optional[Token]) -> None:
    if not (path / "__init__.eiko").exists():
        raise EikoCompilationError(
            "Import path not valid.",
            token=token,
        )


def resolve_from_import(
    import_path: list[str], context: "CompilerContext", dots: list[Token]
) -> Optional[Module]:
    """
    Tries to from import a given path list.
    """
    if dots:
        anchor = context.path.absolute()
        for token in dots:
            anchor = anchor.parent
            test_path_is_module(anchor, token)

        if not import_path:
            import_path.append("__init__")

        return _resolve_from_import(import_path.copy(), anchor, context)

    for _path in PATHS:
        module = _resolve_from_import(import_path.copy(), _path, context)
        if module is not None:
            return module

    return None


def _resolve_from_import(
    import_path: list[str], parent: Path, context: "CompilerContext"
) -> Optional[Module]:
    current = import_path[0]
    import_path.remove(current)

    current_dir = parent / current
    if current_dir.exists() and current_dir.is_dir():
        init_file = current_dir / "__init__.eiko"
        if init_file.exists():
            import_path_copy = import_path.copy()
            import_path_copy.append(current)
            new_context = context.get_cached_context(import_path_copy)
            if new_context is None:
                new_context = context.get_or_set_context(current)
                new_context.set_path(init_file)
            if len(import_path) == 0:
                module = Module(current_dir.stem, init_file, new_context)
                _get_submodules(module)
                return module

            return _resolve_import(import_path, current_dir, new_context)

        return None

    if len(import_path) == 0:
        file_path = parent / (current + ".eiko")
        if file_path.exists():
            new_context = context.get_or_set_context(current)
            module = Module(file_path.stem, file_path, new_context)
            _get_submodules(module)
            return module

    return None


def import_python_code(
    module_path: list[str], eiko_file_path: Path, context: "CompilerContext"
) -> None:
    """
    Resolves and exposes python code that is tagged as eiko-plugins.
    """
    file_path = eiko_file_path.with_suffix(".py")
    if file_path.exists():
        logger.debug(f"Found python plugins file for eiko file: {eiko_file_path}")
        module_name = ".".join(module_path)
        py_module = load_python_code(module_name, file_path)
        for member in getmembers(py_module):
            _obj = member[1]
            if isfunction(_obj) and hasattr(_obj, "eiko_plugin"):
                plugin_name = _obj.__name__
                if hasattr(_obj, "alias"):
                    if _obj.alias is not None:
                        plugin_name = _obj.alias
                logger.debug(f"Importing plugin '{plugin_name}' from {file_path}")
                if context.shallow_get(plugin_name) is None:
                    context.set(
                        plugin_name, _load_plugin(module_name, plugin_name, _obj)
                    )

            if isclass(_obj):
                if issubclass(_obj, Handler) and _obj not in (
                    Handler,
                    CRUDHandler,
                ):
                    logger.debug(
                        f"Importing handler '{_obj.__name__}' from {file_path}"
                    )
                    context.register_handler(_obj)
                elif (
                    issubclass(_obj, EikoBaseModel)
                    and _obj is not EikoBaseModel
                    and (
                        _obj.__module__
                        in (module_name, "eikobot.core.lib." + module_name)
                    )
                ):
                    logger.debug(
                        f"Importing Python Model '{_obj.__name__}' from {file_path}"
                    )
                    context.register_model(_obj)
    else:
        logger.debug(f"Found no python plugins for eiko file: {eiko_file_path}")


def load_python_code(module_name: str, file_path: Path) -> ModuleType:
    """
    Loads Python code from a given path.
    """
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    if spec is not None:
        py_module = importlib.util.module_from_spec(spec)
        if spec.loader is None:
            raise EikoCompilationError(
                f"Failed to import python module {module_name}. (spec.loader is None)"
            )

        spec.loader.exec_module(py_module)
        return py_module

    raise EikoCompilationError(f"Failed to import python module {module_name}.")


def _load_plugin(module: str, name: str, function: Callable) -> PluginDefinition:
    fullargspec = getfullargspec(function)
    annotations = fullargspec.annotations
    return_type = annotations.get("return", "")
    if return_type == "":
        raise EikoCompilationError(
            f"Plugin {module}.{name} return type annotation is missing."
        )
    plugin_definition = PluginDefinition(function, annotations["return"], name, module)

    for arg_name in fullargspec.args:
        arg_annotation = annotations.get(arg_name)
        if arg_annotation is None:
            raise EikoCompilationError(
                f"Plugin '{module}.{name}' has no type annotation for argument '{arg_name}'."
            )

        plugin_definition.add_arg(
            PluginArg(
                arg_name,
                arg_annotation,
            )
        )

    return plugin_definition


def _get_submodules(module: Module) -> None:
    for path in module.path.parent.glob("*"):
        if path.is_dir():
            init_file = path / "__init__.eiko"
            if init_file.exists():
                new_context = module.context.get_or_set_context(path.stem)
                new_module = Module(path.stem, init_file, new_context)
                module.submodules.append(new_module)
                _get_submodules(new_module)

        # Assume this has already been imported
        elif path.name == "__init__.eiko":
            pass

        elif path.suffix == ".eiko":
            module.submodules.append(
                Module(path.stem, path, module.context.get_or_set_context(path.stem))
            )



"""
Typedef definitions are executable functions of sorts.
They can alias types and even put restrictions on them.
"""
from typing import TYPE_CHECKING, Optional, Union

from ...errors import EikoCompilationError
from .._token import Token
from .base_types import (
    BuiltinTypes,
    EikoBaseType,
    EikoBool,
    EikoFloat,
    EikoInt,
    EikoPath,
    EikoStr,
    EikoType,
)

if TYPE_CHECKING:
    from .._parser import ExprAST
    from .context import CompilerContext


class EikoTypeDef(EikoBaseType):
    """Simple custom type definitions."""

    def __init__(
        self,
        name: str,
        super_type: Union[EikoType, "EikoTypeDef"],
        condition: Optional["ExprAST"],
        context: "CompilerContext",
    ) -> None:
        self.name = name
        self.condition = condition
        self.context = context
        self.super = super_type
        if isinstance(super_type, EikoType):
            self.type = EikoType(name, super_type, self)
        else:
            self.type = EikoType(name, super_type.type, self)
        super().__init__(self.type)

    def printable(self, _: str = "") -> str:
        return f"TypeDef '{self.name}' redefining '{self.type.super}'"

    def truthiness(self) -> bool:
        raise NotImplementedError

    def execute(self, arg: EikoBaseType, arg_token: Optional[Token]) -> BuiltinTypes:
        """
        Cast a value to a type and make sure it fits the given condition expression.
        """
        if arg.type.type_check(self.type):
            raise EikoCompilationError(
                f"Type '{self.name}' requires '{self.type.name}' but was passed '{arg.type.name}'.",
                token=arg_token,
            )

        if isinstance(self.super, EikoType):
            base_constructor = self.context.get(self.super.name)
            if base_constructor in (EikoBool, EikoFloat, EikoInt, EikoPath, EikoStr):
                arg = base_constructor(arg.get_value(), self.type)  # type: ignore

        elif isinstance(self.super, EikoTypeDef):
            arg = self.super.execute(arg, arg_token)
            arg.type = EikoType(self.name, arg.type)

        if self.condition is not None:
            condition_context = self.context.get_subcontext(f"{self.name}-typedef")
            condition_context.set("self", arg)
            res = self.condition.compile(condition_context)
            if not isinstance(res, EikoBaseType) or not res.truthiness():
                raise EikoCompilationError(
                    f"Value '{arg.get_value()}' did not meet typedef condition for '{self.name}'.",
                    token=arg_token,
                )

        if isinstance(arg, (EikoBool, EikoFloat, EikoInt, EikoPath, EikoStr)):
            return arg

        raise NotImplementedError



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-statements
from pathlib import Path

import pytest

from eikobot.core.compiler._parser import (
    AssignmentExprAST,
    BinOP,
    BinOpExprAST,
    CallExprAst,
    EnumExprAst,
    FromImportExprAST,
    FStringExprAST,
    FStringLexer,
    IntExprAST,
    ListExprAST,
    Parser,
    ResourceDefinitionAST,
    ResourcePropertyAST,
    StringExprAST,
    TypedefExprAST,
    TypeExprAST,
    UnaryNegExprAST,
    VariableExprAST,
)
from eikobot.core.compiler._token import Index, Token, TokenType
from eikobot.core.compiler.definitions.context import CompilerContext
from eikobot.core.errors import EikoParserError


def test_basic_ops(eiko_basic_ops_file: Path) -> None:
    parser = Parser(eiko_basic_ops_file)
    parse_iter = parser.parse()

    expr_1 = next(parse_iter)
    assert isinstance(expr_1, BinOpExprAST)
    assert expr_1.bin_op == BinOP.ADD
    assert isinstance(expr_1.rhs, IntExprAST)
    assert expr_1.rhs.value == 3
    assert isinstance(expr_1.lhs, BinOpExprAST)
    assert expr_1.lhs.bin_op == BinOP.ADD
    assert isinstance(expr_1.lhs.lhs, IntExprAST)
    assert expr_1.lhs.lhs.value == 1
    assert isinstance(expr_1.lhs.rhs, BinOpExprAST)
    assert expr_1.lhs.rhs.bin_op == BinOP.INT_DIVIDE
    assert isinstance(expr_1.lhs.rhs.lhs, IntExprAST)
    assert expr_1.lhs.rhs.lhs.value == 4
    assert isinstance(expr_1.lhs.rhs.rhs, IntExprAST)
    assert expr_1.lhs.rhs.rhs.value == 2

    expr_2 = next(parse_iter)
    assert isinstance(expr_2, BinOpExprAST)
    assert expr_2.bin_op == BinOP.ADD

    assert isinstance(expr_2.lhs, BinOpExprAST)
    assert expr_2.lhs.bin_op == BinOP.ADD
    assert isinstance(expr_2.lhs.lhs, IntExprAST)
    assert expr_2.lhs.lhs.value == 1
    assert isinstance(expr_2.lhs.rhs, BinOpExprAST)
    assert expr_2.lhs.rhs.bin_op == BinOP.INT_DIVIDE
    assert isinstance(expr_2.lhs.rhs.lhs, IntExprAST)
    assert expr_2.lhs.rhs.lhs.value == 2
    assert isinstance(expr_2.lhs.rhs.rhs, UnaryNegExprAST)
    assert isinstance(expr_2.lhs.rhs.rhs.rhs, BinOpExprAST)
    assert expr_2.lhs.rhs.rhs.rhs.bin_op == BinOP.EXPONENTIATION
    assert isinstance(expr_2.lhs.rhs.rhs.rhs.lhs, IntExprAST)
    assert expr_2.lhs.rhs.rhs.rhs.lhs.value == 3
    assert isinstance(expr_2.lhs.rhs.rhs.rhs.rhs, IntExprAST)
    assert expr_2.lhs.rhs.rhs.rhs.rhs.value == 2

    assert isinstance(expr_2.rhs, BinOpExprAST)
    assert expr_2.rhs.bin_op == BinOP.MULTIPLY
    assert isinstance(expr_2.rhs.lhs, UnaryNegExprAST)
    assert isinstance(expr_2.rhs.lhs.rhs, IntExprAST)
    assert expr_2.rhs.lhs.rhs.value == 7
    assert isinstance(expr_2.rhs.rhs, IntExprAST)
    assert expr_2.rhs.rhs.value == 2

    expr_3 = next(parse_iter)
    assert isinstance(expr_3, BinOpExprAST)
    assert expr_3.bin_op == BinOP.ADD
    assert isinstance(expr_3.rhs, UnaryNegExprAST)
    assert isinstance(expr_3.rhs.rhs, ListExprAST)
    expr_3_rhs_rhs = expr_3.rhs.rhs.elements[0]
    assert isinstance(expr_3_rhs_rhs, BinOpExprAST)
    assert expr_3_rhs_rhs.bin_op == BinOP.MULTIPLY
    assert isinstance(expr_3_rhs_rhs.rhs, IntExprAST)
    assert expr_3_rhs_rhs.rhs.value == 2
    assert isinstance(expr_3_rhs_rhs.lhs, IntExprAST)
    assert expr_3_rhs_rhs.lhs.value == 7
    assert isinstance(expr_3.lhs, BinOpExprAST)
    assert expr_3.lhs.bin_op == BinOP.EXPONENTIATION
    assert isinstance(expr_3.lhs.rhs, IntExprAST)
    assert expr_3.lhs.rhs.value == 2
    assert isinstance(expr_3.lhs.lhs, ListExprAST)
    expr_3_lhs_lhs = expr_3.lhs.lhs.elements[0]
    assert isinstance(expr_3_lhs_lhs, BinOpExprAST)
    assert expr_3_lhs_lhs.bin_op == BinOP.INT_DIVIDE
    assert isinstance(expr_3_lhs_lhs.lhs, ListExprAST)
    expr_3_lhs_lhs_lhs = expr_3_lhs_lhs.lhs.elements[0]
    assert isinstance(expr_3_lhs_lhs_lhs, BinOpExprAST)
    assert expr_3_lhs_lhs_lhs.bin_op == BinOP.ADD
    assert isinstance(expr_3_lhs_lhs_lhs.rhs, IntExprAST)
    assert expr_3_lhs_lhs_lhs.rhs.value == 2
    assert isinstance(expr_3_lhs_lhs_lhs.lhs, IntExprAST)
    assert expr_3_lhs_lhs_lhs.lhs.value == 1
    assert isinstance(expr_3_lhs_lhs.rhs, UnaryNegExprAST)
    assert isinstance(expr_3_lhs_lhs.rhs.rhs, IntExprAST)
    assert expr_3_lhs_lhs.rhs.rhs.value == 3

    expr_4 = next(parse_iter)
    assert isinstance(expr_4, BinOpExprAST)
    assert expr_4.bin_op == BinOP.ADD
    assert isinstance(expr_4.rhs, StringExprAST)
    assert expr_4.rhs.value == "string 2"
    assert isinstance(expr_4.lhs, BinOpExprAST)
    assert expr_4.lhs.bin_op == BinOP.ADD
    assert isinstance(expr_4.lhs.lhs, StringExprAST)
    assert expr_4.lhs.lhs.value == "string 1"
    assert isinstance(expr_4.lhs.rhs, StringExprAST)
    assert expr_4.lhs.rhs.value == " + "

    expr_5 = next(parse_iter)
    assert isinstance(expr_5, BinOpExprAST)
    assert expr_5.bin_op == BinOP.SUBTRACT
    assert isinstance(expr_5.rhs, IntExprAST)
    assert expr_5.rhs.value == 4
    assert isinstance(expr_5.lhs, BinOpExprAST)
    assert expr_5.lhs.bin_op == BinOP.SUBTRACT
    assert isinstance(expr_5.lhs.lhs, IntExprAST)
    assert expr_5.lhs.lhs.value == 4
    assert isinstance(expr_5.lhs.rhs, IntExprAST)
    assert expr_5.lhs.rhs.value == 4

    expr_6 = next(parse_iter)
    assert isinstance(expr_6, BinOpExprAST)
    assert expr_6.bin_op == BinOP.MULTIPLY
    assert isinstance(expr_6.lhs, StringExprAST)
    assert expr_6.lhs.value == "ha"
    assert isinstance(expr_6.rhs, IntExprAST)
    assert expr_6.rhs.value == 3

    expr_7 = next(parse_iter)
    assert isinstance(expr_7, StringExprAST)
    assert expr_7.value == "auto concat string"

    with pytest.raises(StopIteration):
        next(parse_iter)


def test_parse_resource(eiko_file_1: Path) -> None:
    parser = Parser(eiko_file_1)
    parse_iter = parser.parse()

    expr_1 = next(parse_iter)
    assert isinstance(expr_1, ResourceDefinitionAST)
    assert len(expr_1.properties) == 2
    prop_1 = expr_1.properties["ip"]
    assert isinstance(prop_1, ResourcePropertyAST)
    assert prop_1.name == "ip"
    assert isinstance(prop_1.expr, VariableExprAST)
    assert isinstance(prop_1.expr.type_expr, TypeExprAST)

    prop_2 = expr_1.properties["ip_2"]
    assert isinstance(prop_2.expr, VariableExprAST)
    assert isinstance(prop_2.expr.type_expr, TypeExprAST)

    var_1 = next(parse_iter)
    assert isinstance(var_1, AssignmentExprAST)
    assert isinstance(var_1.lhs, VariableExprAST)
    assert var_1.lhs.identifier == "test_1"
    assert isinstance(var_1.rhs, CallExprAst)
    assert var_1.rhs.identifier == "Test"
    assert len(var_1.rhs.args.elements) == 2
    assert isinstance(var_1.rhs.args.elements[0], StringExprAST)
    assert var_1.rhs.args.elements[0].value == "192.168.0.1"
    assert isinstance(var_1.rhs.args.elements[1], StringExprAST)
    assert var_1.rhs.args.elements[1].value == "192.168.1.1"


def test_f_string_lexer() -> None:
    lexer = FStringLexer(
        Token(
            TokenType.F_STRING,
            "This is an f-string test: {3 + 3}",
            Index(0, 0, Path("test_parser.py")),
        )
    )
    token_0 = lexer.next_token()
    assert token_0 == Token(TokenType.INDENT, "", Index(0, 27, lexer.file_path))
    token_1 = lexer.next_token()
    assert token_1 == Token(TokenType.INTEGER, "3", Index(0, 27, lexer.file_path))
    token_2 = lexer.next_token()
    assert token_2 == Token(TokenType.ARITHMETIC_OP, "+", Index(0, 29, lexer.file_path))
    token_3 = lexer.next_token()
    assert token_3 == Token(TokenType.INTEGER, "3", Index(0, 31, lexer.file_path))
    token_4 = lexer.next_token()
    assert token_4 == Token(TokenType.EOF, "EOF", Index(1, 0, lexer.file_path))

    lexer = FStringLexer(
        Token(
            TokenType.F_STRING,
            "This is an f-string test: {3",
            Index(0, 0, Path("test_parser.py")),
        )
    )

    token_0 = lexer.next_token()
    assert token_0 == Token(TokenType.INDENT, "", Index(0, 27, lexer.file_path))
    token_1 = lexer.next_token()
    assert token_1 == Token(TokenType.INTEGER, "3", Index(0, 27, lexer.file_path))

    with pytest.raises(EikoParserError):
        token_2 = lexer.next_token()


def test_f_string_parser(eiko_f_string_file: Path) -> None:
    f_string_expr = FStringExprAST(
        Token(
            TokenType.F_STRING,
            "This is an f-string test: {3 + 3}, {4 + 4}",
            Index(0, 0, eiko_f_string_file),
        )
    )
    expr_1 = f_string_expr.expressions.get("{3 + 3}")
    assert isinstance(expr_1, BinOpExprAST)
    expr_2 = f_string_expr.expressions.get("{3 + 3}")
    assert isinstance(expr_2, BinOpExprAST)

    compiled_f_string = f_string_expr.compile(CompilerContext("f-string-test", {}))
    assert compiled_f_string.value == f"This is an f-string test: {3 + 3}, {4 + 4}"


def test_parse_typedef(eiko_typedef: Path) -> None:
    parser = Parser(eiko_typedef)
    parse_iter = parser.parse()

    expr_1 = next(parse_iter)
    assert isinstance(expr_1, FromImportExprAST)

    expr_2 = next(parse_iter)
    assert isinstance(expr_2, TypedefExprAST)
    assert expr_2.name == "string_alias"
    assert expr_2.super_type_expr.token.content == "str"
    assert expr_2.condition is None

    expr_3 = next(parse_iter)
    assert isinstance(expr_3, TypedefExprAST)
    assert expr_3.name == "IPv4Address"
    assert expr_3.super_type_expr.token.content == "str"
    assert isinstance(expr_3.condition, CallExprAst)


def test_parse_enum(eiko_enum_file: Path) -> None:
    parser = Parser(eiko_enum_file)
    parse_iter = parser.parse()

    expr_1 = next(parse_iter)
    assert isinstance(expr_1, EnumExprAst)

    expr_2 = next(parse_iter)
    assert isinstance(expr_2, ResourceDefinitionAST)



"""
The entrypoint to the client application.
Schould only contain things related to the client cli.
"""
import asyncio
import datetime
import signal
import subprocess
import sys
import time
import traceback
from pathlib import Path

import click

from . import VERSION
from .core import logger, package_manager
from .core.compiler import Compiler
from .core.compiler.lexer import Token
from .core.deployer import Deployer
from .core.errors import EikoError, EikoPluginError
from .core.exporter import Exporter
from .core.project import PROJECT_SETTINGS, init_project


@click.group()
@click.option("--debug", is_flag=True)
@click.version_option(VERSION)
def cli(debug: bool = False) -> None:
    """
    The Eikobot CLI allows for compilation
    and exporting of eiko files.
    """
    log_level = logger.LogLevel.INFO
    if debug:
        log_level = logger.LogLevel.DEBUG
    logger.init(log_level=log_level)


@cli.command(name="compile")
@click.option("-f", "--file", "file", help="Path to entrypoint file.")
@click.option(
    "--output-model",
    is_flag=True,
    help="Outputs a human readable version of compiler context.",
)
@click.option(
    "--enable-plugin-stacktrace",
    is_flag=True,
    help="Outputs a plugins stacktrace if it raises an exception.",
)
def compile_cmd(
    file: str | None, output_model: bool = False, enable_plugin_stacktrace: bool = False
) -> None:
    """
    Compile an eikobot file.
    """
    _compile(file, output_model, enable_plugin_stacktrace)


def _get_file_path(file: str | None) -> Path:
    if file is None:
        if PROJECT_SETTINGS.entry_point is None:
            file = ""
            while file == "":
                file = input("File: ")
        else:
            file = PROJECT_SETTINGS.entry_point

    return Path(file)


def _compile(
    file: str | None, output_model: bool = False, enable_plugin_stacktrace: bool = False
) -> Compiler:
    file_path = _get_file_path(file)

    start = time.time()
    pc_start = time.process_time()
    compiler = Compiler()

    if not file_path.exists():
        logger.error(f"No such file: {file_path}")
        sys.exit(1)

    logger.info(f"Compiling {file_path}")
    try:
        compiler.compile(file_path)
    except EikoError as e:
        logger.error(str(e))

        if e.index is not None:
            logger.print_error_trace(e.index)

        if isinstance(e, EikoPluginError):
            if e.python_exception is not None:
                logger.error("Python error: " + str(e.python_exception))
                if enable_plugin_stacktrace:
                    logger.error("Python plugin stacktrace (ignore the first line):")
                    traceback.print_exception(e.python_exception)
                else:
                    logger.error(
                        "To view the python plugin stacktrace, use '--enable-plugin-stacktrace'"
                    )
        sys.exit(1)

    except NotImplementedError as e:
        logger.error("PANIC!! Compiler ran in to an unexpected error!")
        logger.error("Please report this error on github.")
        try:
            token = e.args[0]
            if isinstance(token, Token):
                logger.error("Got stuck here:")
                logger.print_error_trace(token.index)
        except IndexError:
            pass
        sys.exit(1)

    if output_model:
        logger.info("resulting model context:")
        print(compiler.context)

    pc_time_taken = time.process_time() - pc_start
    time_taken = time.time() - start
    pc_time_taken_formatted = str(datetime.timedelta(seconds=pc_time_taken))
    time_taken_formatted = str(datetime.timedelta(seconds=time_taken))
    logger.info(
        f"Compiled in {time_taken_formatted} 🤖 "
        f"(Process time: {pc_time_taken_formatted})"
    )

    return compiler


@cli.command()
@click.option("-f", "--file", "file", help="Path to entrypoint file.")
@click.option(
    "--enable-plugin-stacktrace",
    is_flag=True,
    help="Outputs a plugins stacktrace if it raises an exception.",
)
@click.option(
    "--dry-run",
    is_flag=True,
    help="Does a dry run of all the tasks in a given model.",
)
@click.option(
    "--force",
    is_flag=True,
    help="Overwrites ignores any dry-run parameters",
)
def deploy(
    file: str | None,
    enable_plugin_stacktrace: bool = False,
    dry_run: bool = False,
    force: bool = False,
) -> None:
    """
    Compile, export and deploy a model from a given file.
    """
    start = time.time()
    compiler = _compile(file, False, enable_plugin_stacktrace)
    logger.info("Exporting model.")
    try:
        exporter = Exporter()
        exporter.export_from_context(compiler.context)
        deployer = Deployer()
        if (dry_run or PROJECT_SETTINGS.dry_run) and not force:
            logger.info("Performing dry run.")
            asyncio.run(deployer.dry_run(exporter))
        else:
            logger.info("Deploying model.")
            asyncio.run(deployer.deploy(exporter, log_progress=True))
    except EikoError as e:
        logger.error(str(e))

        if e.index is not None:
            logger.print_error_trace(e.index)
    else:
        if deployer.failed:
            logger.error(
                "Failed to deploy model. "
                f"({deployer.progress.done} out of {deployer.progress.total} tasks done)"
            )
        else:
            time_taken = time.time() - start
            time_taken_formatted = str(datetime.timedelta(seconds=time_taken))
            logger.info(f"Deployed in {time_taken_formatted} 🤖")


@cli.group()
def package() -> None:
    pass


@package.command(name="build")
def build_pkg() -> None:
    """Builds a package in the cwd, using an eiko.toml file."""
    try:
        package_manager.build_pkg()
    except EikoError as e:
        logger.error(str(e))


@package.command(name="install")
@click.argument("target")
def install_pkg(target: str) -> None:
    """
    Install a package from different sources.
    """
    if target == ".":
        requires: list[str] = []
        _pkg_data_path = Path("eiko.toml")
        if _pkg_data_path.exists():
            try:
                pkg_data = package_manager.read_pkg_toml(_pkg_data_path)
                requires.extend(pkg_data.eikobot_requires)
                if pkg_data.python_requires:
                    package_manager.install_py_deps(pkg_data.python_requires)
            except EikoError:
                pass

        if len(requires) > 0:
            asyncio.run(package_manager.install_pkgs(requires))
        elif not pkg_data.python_requires:
            logger.error("No requirements found.")

    else:
        asyncio.run(_install_pkg(target))


async def _install_pkg(target: str) -> None:
    try:
        await package_manager.install_pkg(target)
    except EikoError as e:
        logger.error(str(e))


@package.command(name="list")
def list_pkg() -> None:
    """
    Lists all installed packages.
    """
    packages = package_manager.get_pkg_index()
    for pkg in packages.values():
        print(f"{pkg.name}=={pkg.version}")


@package.group()
def release() -> None:
    pass


@release.command()
def github() -> None:
    """
    Creates a release
    """
    try:
        asyncio.run(package_manager.create_github_release())
    except EikoError as e:
        logger.error(str(e))


@package.command(name="uninstall")
@click.argument("name")
def uninstall_pkg(name: str) -> None:
    """
    Performs all required steps to delete a package.
    """
    try:
        package_manager.uninstall_pkg(name)
    except EikoError as e:
        logger.error(str(e))


@cli.group()
def project() -> None:
    pass


@project.command()
def init() -> None:
    """
    Initialize an existing project.
    """
    if PROJECT_SETTINGS.exists:
        asyncio.run(init_project())
    else:
        logger.error("CWD does not have an eiko.toml file.")


def _run_wrapped() -> None:
    # There is a problem with relative imports in python code
    # when we import python code from eiko modules/packages at runtime.
    # This bug only happens if we are not pythons main entrypoint.
    # In other words this is a pretty ugly hack, and I do not know
    # any otehr fixes at this point.
    with subprocess.Popen(
        [
            "python",
            "-m",
            "eikobot",
            *sys.argv[1:],
        ],
    ) as _process:
        try:
            _process.wait()
        except KeyboardInterrupt:
            _process.send_signal(signal.SIGINT)

        sys.exit(_process.returncode)


def main() -> None:
    """
    Python entrypoint function that does some housekeeping.
    """
    if __name__ == "__main__":
        cli()
    else:
        _run_wrapped()


if __name__ == "__main__":
    main()



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-statements
from pathlib import Path

from eikobot.core.compiler import Compiler
from eikobot.core.compiler.definitions.base_types import (
    EikoBool,
    EikoDict,
    EikoInt,
    EikoList,
    EikoNone,
    EikoStr,
)


def test_optional(tmp_eiko_file: Path) -> None:
    compiler = Compiler()
    with open(tmp_eiko_file, "w", encoding="utf-8") as f:
        f.write('a: Optional[str] = "test_string"')

    compiler.compile(tmp_eiko_file)

    var_a = compiler.context.get("a")
    assert isinstance(var_a, EikoStr)
    assert var_a.value == "test_string"

    with open(tmp_eiko_file, "w", encoding="utf-8") as f:
        f.write("a: Optional[str] = None")

    compiler.reset()
    compiler.compile(tmp_eiko_file)

    var_a = compiler.context.get("a")
    assert isinstance(var_a, EikoNone)


def test_list(eiko_list_file: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_list_file)

    var_test_list = compiler.context.get("test_list")
    assert isinstance(var_test_list, EikoList)
    assert len(var_test_list.elements) == 4

    element_0 = var_test_list.elements[0]
    assert isinstance(element_0, EikoStr)
    assert element_0.value == "string_1"

    element_1 = var_test_list.elements[1]
    assert isinstance(element_1, EikoStr)
    assert element_1.value == "string_2"

    element_2 = var_test_list.elements[2]
    assert isinstance(element_2, EikoInt)
    assert element_2.value == 3

    element_3 = var_test_list.elements[3]
    assert isinstance(element_3, EikoBool)
    assert element_3.value is True

    var_a = compiler.context.get("a")
    assert isinstance(var_a, EikoStr)
    assert var_a.value == "string_2"

    var_b = compiler.context.get("b")
    assert isinstance(var_b, EikoInt)
    assert var_b.value == 3


def test_dict(eiko_dict_file: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_dict_file)

    var_test_dict = compiler.context.get("test_dict")
    assert isinstance(var_test_dict, EikoDict)
    assert len(var_test_dict.elements) == 4

    value_1 = var_test_dict.elements.get("key_1")
    assert isinstance(value_1, EikoStr)
    assert value_1.value == "string_1"

    value_2 = var_test_dict.elements.get("key_2")
    assert isinstance(value_2, EikoStr)
    assert value_2.value == "string_2"

    value_3 = var_test_dict.elements.get("key_3")
    assert isinstance(value_3, EikoInt)
    assert value_3.value == 3

    var_a = compiler.context.get("a")
    assert isinstance(var_a, EikoStr)
    assert var_a.value == "string_2"

    value_4 = var_test_dict.elements.get("key_4")
    assert isinstance(value_4, EikoBool)
    assert value_4.value is True



"""
Eikobot plugins expose python fucntions
as callables directly to the Eikobot runtime.
"""
from typing import Callable, Optional, Union

from . import human_readable, machine_readable
from .compiler.definitions.base_types import EikoBaseType

EikoPluginTyping = Callable[..., Union[None, bool, float, int, str, EikoBaseType]]


def eiko_plugin(alias: Optional[str] = None) -> Callable:
    """Tags a function as an Eikobot plugin with alias."""

    def _eiko_plugin(function: EikoPluginTyping) -> EikoPluginTyping:
        """Tags a function as an Eikobot plugin."""
        function.eiko_plugin = True  # type: ignore
        function.alias = alias  # type: ignore

        return function

    return _eiko_plugin


class EikoPluginException(Exception):
    """
    If something goes wrong inside a plugin, this error should be thrown.
    """


@eiko_plugin("type")
def eiko_type(
    obj: EikoBaseType,
) -> str:
    """
    Returns the type of a an eiko object as a string.
    """
    return obj.type.name


@eiko_plugin("human_readable")
def eiko_human_readable(number: int) -> str:
    """
    plugin wrapper for core function human_readable
    """
    return human_readable(number)


@eiko_plugin("machine_readable")
def eiko_machine_readable(number: str) -> int:
    """
    plugin wrapper for core function machine_readable
    """
    return machine_readable(number)



"""
The regex module contains bindings to pythons re module.
"""
import re

from eikobot.core.plugin import eiko_plugin


@eiko_plugin()
def match(regex: str, string: str) -> bool:
    """Returns a True if python re.match returns a result."""
    if re.match(regex, string) is not None:
        return True

    return False



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-statements
from pathlib import Path

from eikobot.core.compiler import Compiler
from eikobot.core.compiler.decorator import index_decorator
from eikobot.core.compiler.definitions.base_types import EikoResource


def test_decorator(eiko_decorator_file: Path) -> None:
    compiler = Compiler()
    compiler.compile(eiko_decorator_file)

    _index_decorator = compiler.context.get("index")
    assert _index_decorator is index_decorator

    _command_res = compiler.context.get("cmd")
    assert isinstance(_command_res, EikoResource)
    assert _command_res.index() == "TCommand-127.0.0.1-echo test"



"""
The setup.py file is widely used format.
Contains all the logic for making the packages.
"""
from pathlib import Path

import setuptools

this_directory = Path(__file__).parent
with open(this_directory / "README.md", encoding="utf-8") as f:
    long_description = f.read()

VERSION = "0.7.3"
REQUIRES = [
    "aiohttp==3.8.5",
    "asyncssh==2.13.1",
    "click==8.1.3",
    "colorama==0.4.6",
    "jinja2==3.1.2",
    "packaging==23.1",
    "pydantic==1.10.4",
]

setuptools.setup(
    name="eikobot",
    version=VERSION,
    description="The eikobot desired state engine.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="kazaamjt",
    author_email="kazaamjt@gmail.com",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Development Status :: 4 - Beta",
    ],
    packages=setuptools.find_packages(),
    package_data={"": ["**/*.eiko", "**/*.py", "py.typed"]},
    include_package_data=True,
    python_requires=">=3.11",
    install_requires=REQUIRES,
    entry_points={"console_scripts": ["eikobot=eikobot.__main__:main"]},
)



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
# pylint: disable=too-many-statements
from pathlib import Path

import pytest

from eikobot.core.deployer import Deployer


@pytest.mark.asyncio
async def test_promises(eiko_promises_file: Path) -> None:
    deployer = Deployer()
    await deployer.deploy_from_file(eiko_promises_file)



"""
Context hold variables, classes and more.
Used both by files/modules and fucntions.
"""
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Type, Union

from ... import logger
from ...errors import EikoCompilationError, EikoInternalError
from ...handlers import Handler
from ...plugin import eiko_type, human_readable, machine_readable
from .._token import Token
from ..decorator import index_decorator
from ..importlib import _load_plugin, import_python_code
from ._resource import EikoResourceDefinition
from .base_model import EikoBaseModel
from .base_types import (
    BuiltinTypes,
    EikoBaseType,
    EikoBool,
    EikoDictType,
    EikoFloat,
    EikoInt,
    EikoListType,
    EikoOptional,
    EikoPath,
    EikoStr,
    EikoType,
    EikoUnion,
    EikoUnset,
    eiko_none_object,
)
from .typedef import EikoTypeDef

if TYPE_CHECKING:
    from .._parser import Parser

_StorableTypes = Union[
    EikoBaseType, EikoResourceDefinition, Type[EikoBaseType], EikoType, Type[EikoType]
]
_builtins: dict[str, _StorableTypes] = {
    "int": EikoInt,
    "float": EikoFloat,
    "bool": EikoBool,
    "str": EikoStr,
    "Path": EikoPath,
    "None": eiko_none_object,
    "Union": EikoUnion,
    "Optional": EikoOptional,
    "list": EikoListType,
    "dict": EikoDictType,
    "index": index_decorator,
    "type": _load_plugin("", "type", eiko_type),
    "human_readable": _load_plugin("", "human_readable", human_readable),
    "machine_readable": _load_plugin("", "machine_readable", machine_readable),
}


@dataclass
class LazyLoadModule:
    """A lazyLoadModule is meant to only be compiled when it is directly called."""

    context: "CompilerContext"
    parser: "Parser"
    import_path: list[str]

    def compile(self) -> "CompilerContext":
        """Imports plugins and compiles eiko code so the module can be used."""
        if not self.context.compiled:
            logger.debug(f"Importing module '{'.'.join(self.import_path)}'")
            for expr in self.parser.parse():
                expr.compile(self.context)

            import_python_code(
                self.import_path,
                self.parser.lexer.file_path,
                self.context,
            )
            self.context.flag_as_compiled()
        return self.context


class CompilerContext:
    """
    Context hold variables, classes and more.
    Used both by files/modules and functions.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name: str,
        context_cache: dict[str, "CompilerContext"],
        super_scope: Optional["CompilerContext"] = None,
        super_module: Optional["CompilerContext"] = None,
        is_root: bool = False,
    ) -> None:
        self.name = name
        self.storage: dict[
            str,
            Union[_StorableTypes, "CompilerContext", LazyLoadModule, EikoUnset, None],
        ] = {}
        self.path: Path
        self.type = EikoType("eiko_internal_context")
        self.super = super_scope
        self.is_root = is_root

        if super_module is not None and super_module.is_root:
            super_module = None
        self.super_module = super_module
        self._context_cache: dict[str, CompilerContext] = context_cache

        self.compiled = False
        self.handlers: dict[str, Type[Handler]] = {}
        self.models: dict[str, Type[EikoBaseModel]] = {}
        self.orphans: list[EikoBaseType] = []

        self.global_id_list: list[str]
        if self.super is not None:
            self.global_id_list = self.super.global_id_list
        elif self.super_module is not None:
            self.global_id_list = self.super_module.global_id_list
        else:
            self.global_id_list = []

    @classmethod
    def convert(cls, _: "BuiltinTypes") -> "EikoBaseType":
        raise ValueError

    def flag_as_compiled(self) -> None:
        self.compiled = True

    def set_path(self, path: Path) -> None:
        self.path = path
        self.storage["__file__"] = EikoPath(path)

    def __repr__(self, indent: str = "") -> str:
        return_str = indent + f"Context '{self.name}': " + "{\n"
        extra_indent = indent + "    "
        for key, value in self.storage.items():
            if isinstance(value, CompilerContext):
                return_str += value.__repr__(extra_indent)
            elif isinstance(value, LazyLoadModule):
                pass
            elif isinstance(value, EikoResourceDefinition):
                return_str += value.printable(extra_indent)
            elif isinstance(value, EikoBaseType):
                extra_extra_indent = extra_indent + "    "
                return_str += f"{extra_indent}var '{key}': "
                return_str += value.printable(extra_extra_indent)
                return_str += "\n"
            else:
                return_str += f"{extra_indent}{key}: {value}\n"

        return_str += indent + "}\n"
        return return_str

    def get(
        self, name: str, token: Optional[Token] = None
    ) -> Union[_StorableTypes, "CompilerContext", None]:
        """Get a value from this context or a super context."""
        value = self.storage.get(name)

        if isinstance(value, LazyLoadModule):
            value = value.compile()
            self.storage[name] = value
        elif value is None and self.super is not None:
            value = self.super.get(name, token)

        if value is None:
            value = _builtins.get(name)

        if isinstance(value, EikoUnset):
            raise EikoCompilationError(
                "Variable accessed before having been assignend a value.",
                token=token,
            )

        return value

    def shallow_get(
        self, name: str
    ) -> Union[_StorableTypes, "CompilerContext", EikoUnset, None]:
        """
        Shallow get only gets builtins and values local to the current scope.
        It is primarily for use by 'Set'.
        """
        value = self.storage.get(name)

        if isinstance(value, LazyLoadModule):
            value = value.compile()
        elif value is None:
            value = _builtins.get(name)

        return value

    def set(
        self,
        name: str,
        value: Union[_StorableTypes, "CompilerContext", EikoUnset],
        token: Optional[Token] = None,
    ) -> None:
        """Set a value. Throws an error if it's already set."""
        prev_value = self.shallow_get(name)
        if isinstance(prev_value, EikoUnset):
            if (
                prev_value.type.inverse_type_check(value.type)
                and prev_value.type.name != value.type.name
            ):
                constr = self.get(prev_value.type.name)
                if isinstance(constr, EikoTypeDef):
                    if isinstance(value, EikoBaseType):
                        value = constr.execute(value, token)
                    else:
                        raise EikoInternalError(
                            "Something went wrong trying to coerce a type to it's typedef.",
                            token=token,
                        )

            if not prev_value.type.type_check(value.type):
                raise EikoCompilationError(
                    f"Tried to assign value of type {value.type} "
                    f"to a variable declared as type {prev_value.type}.",
                    token=token,
                )

        elif prev_value is not None:
            raise EikoCompilationError(
                f"Illegal operation: Tried to reassign '{name}'.",
                token=token,
            )

        self.storage[name] = value
        self._connect_handler(name)
        self._connect_model(name)

    def get_or_set_context(
        self, name: str, token: Optional[Token] = None
    ) -> "CompilerContext":
        """
        Either retrieve a context or create it if it doesn't exist.
        """
        context = self.get(name)
        if isinstance(context, CompilerContext):
            return context

        if context is None:
            new_context = CompilerContext(name, self._context_cache, super_module=self)
            self.set(name, new_context)
            if self.is_root:
                self.add_tl_context(name, new_context)
            return new_context

        raise EikoCompilationError(
            f'Illegal operation: Tried to reassign "{name}".',
            token=token,
        )

    def get_subcontext(self, name: str) -> "CompilerContext":
        """
        Creates a new context, with the given context as its super context,
        but does not store the new context in the super context.
        """
        return CompilerContext(name, self._context_cache, self)

    def register_handler(self, handler: Type[Handler]) -> None:
        """Adds a handler to the context for later retrieval."""
        try:
            prev_handler = self.handlers.get(handler.__eiko_resource__)
        except AttributeError as e:
            raise EikoCompilationError(
                f"Handler '{handler.__name__}' requires an `__eiko_resource__` field "
                "to link to its resource."
            ) from e

        if prev_handler is not None:
            raise EikoCompilationError(
                f"A handler for resource type '{handler.__eiko_resource__}' was already registered."
            )

        self.handlers[handler.__eiko_resource__] = handler
        self._connect_handler(handler.__eiko_resource__)

    def _connect_handler(self, name: str) -> None:
        handler = self.handlers.get(name)
        if handler is None:
            return

        resource = self.storage.get(name)
        if resource is None:
            return

        if not isinstance(resource, EikoResourceDefinition):
            raise EikoCompilationError(
                f"Tried to connect a handler to '{name}', which is not a resource definition."
            )

        logger.debug(
            f"Linking handler {handler} to resource '{self.get_import_name(include_main=True)}.{name}'"
        )
        resource.handler = handler

    def get_import_name(self, include_main: bool = False) -> str:
        """Constructs a name based on inherited contexts."""
        name = ""
        if self.name == "__main__" and not include_main:
            return name

        if self.super_module is not None:
            super_name = self.super_module.get_import_name()
            if super_name != "":
                name += super_name + "."

        return name + self.name

    def register_model(self, model: Type[EikoBaseModel]) -> None:
        """Adds a model to the context for later retrieval."""
        res_name = model.get_resource_name()
        prev_model = self.models.get(res_name)

        if prev_model is not None:
            raise EikoCompilationError(
                f"A Python model for resource type '{res_name}' was already registered."
            )

        self.models[res_name] = model
        self._connect_model(res_name)

    def _connect_model(self, name: str) -> None:
        model = self.models.get(name)
        if model is None:
            return

        resource_cls = self.storage.get(name)
        if resource_cls is None:
            return

        if not isinstance(resource_cls, EikoResourceDefinition):
            raise EikoCompilationError(
                f"Tried to connect a Python model to '{name}', which is not a resource definition."
            )

        logger.debug(
            f"Linking model {model} to resource '{self.get_import_name(include_main=True)}.{name}'"
        )
        model.link(resource_cls)

    def get_top_level_context(self) -> "CompilerContext":
        """Returns the recursivly retrieved top level parent context/module."""
        if self.super_module is not None:
            return self.super_module.get_top_level_context()

        return self

    def get_cached_context(self, import_path: list[str]) -> Optional["CompilerContext"]:
        """Checks to see if a given context already exists."""
        context = self._context_cache.get(import_path[0])
        if context is None:
            return None

        for name in import_path[1:]:
            _context = context.get(name)
            if isinstance(_context, CompilerContext):
                context = _context
            elif isinstance(_context, LazyLoadModule):
                context = _context.context
            else:
                break

        if isinstance(context, LazyLoadModule):
            context = context.compile()

        return context

    def add_tl_context(self, name: str, context: "CompilerContext") -> None:
        """Add a top level context to the cache."""
        self._context_cache[name] = context


StorableTypes = Union[_StorableTypes, "CompilerContext"]
